# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from fractions import Fraction
import sys
import re

from decimal import Decimal, localcontext
from warnings import warn

import unified_planning as up
import unified_planning.environment
import unified_planning.model.walkers as walkers
from unified_planning.model import (
    InstantaneousAction,
    DurativeAction,
    Fluent,
    Parameter,
    Problem,
    Object,
)
from unified_planning.exceptions import (
    UPTypeError,
    UPProblemDefinitionError,
    UPException,
)
from unified_planning.model.types import _UserType
from typing import Callable, Dict, IO, List, Optional, Set, Union, cast
from io import StringIO
from functools import reduce

PDDL_KEYWORDS = {
    "define",
    "domain",
    "requirements",
    "types",
    "constants",
    "atomic",
    "predicates",
    "problem",
    "atomic",
    "constraints",
    "either",
    "number",
    "action",
    "parameters",
    "precondition",
    "effect",
    "and",
    "forall",
    "preference",
    "or",
    "not",
    "imply",
    "exists",
    "scale-up",
    "scale-down",
    "increase",
    "decrease",
    "durative-action",
    "duration",
    "condition",
    "at",
    "over",
    "start",
    "end",
    "all",
    "derived",
    "objects",
    "init",
    "goal",
    "when",
    "decrease",
    "always",
    "sometime",
    "within",
    "at-most-once",
    "sometime-after",
    "sometime-before",
    "always-within",
    "hold-during",
    "hold-after",
    "metric",
    "minimize",
    "maximize",
    "total-time",
    "is-violated",
    "strips",
    "negative-preconditions",
    "typing",
    "disjunctive-preconditions",
    "equality",
    "existential-preconditions",
    "universal-preconditions",
    "quantified-preconditions",
    "conditional-effects",
    "fluents",
    "adl",
    "durative-actions",
    "derived-predicates",
    "timed-initial-literals",
    "preferences",
}

# The following map is used to mangle the invalid names by their class.
INITIAL_LETTER: Dict[type, str] = {
    InstantaneousAction: "a",
    DurativeAction: "a",
    Fluent: "f",
    Parameter: "p",
    Problem: "p",
    Object: "o",
}


class ObjectsExtractor(walkers.DagWalker):
    """Returns the object instances appearing in the expression."""

    def __init__(self):
        walkers.dag.DagWalker.__init__(self)

    def get(self, expression: "up.model.FNode") -> Dict[_UserType, Set[Object]]:
        """Returns all the free vars of the given expression."""
        return self.walk(expression)

    def walk_object_exp(
        self, expression: "up.model.FNode", args: List[Dict[_UserType, Set[Object]]]
    ) -> Dict[_UserType, Set[Object]]:
        res: Dict[_UserType, Set[Object]] = {}
        for a in args:
            _update_domain_objects(res, a)
        obj = expression.object()
        assert obj.type.is_user_type()
        res.setdefault(cast(_UserType, obj.type), set()).add(obj)
        return res

    @walkers.handles(
        set(up.model.OperatorKind).difference((up.model.OperatorKind.OBJECT_EXP,))
    )
    def walk_all_types(
        self, expression: "up.model.FNode", args: List[Dict[_UserType, Set[Object]]]
    ) -> Dict[_UserType, Set[Object]]:
        res: Dict[_UserType, Set[Object]] = {}
        for a in args:
            _update_domain_objects(res, a)
        return res


class ConverterToPDDLString(walkers.DagWalker):
    """Expression converter to a PDDL string."""

    DECIMAL_PRECISION = 10  # Number of decimal places to print real constants

    def __init__(
        self,
        env: "up.environment.Environment",
        get_mangled_name: Callable[
            [
                Union[
                    "up.model.Type",
                    "up.model.Action",
                    "up.model.Fluent",
                    "up.model.Object",
                ]
            ],
            str,
        ],
    ):
        walkers.DagWalker.__init__(self)
        self.get_mangled_name = get_mangled_name
        self.simplifier = env.simplifier

    def convert(self, expression):
        """Converts the given expression to a PDDL string."""
        return self.walk(self.simplifier.simplify(expression))

    def walk_exists(self, expression, args):
        assert len(args) == 1
        vars_string_list = [
            f"{self.get_mangled_name(v)} - {self.get_mangled_name(v.type)}"
            for v in expression.variables()
        ]
        return f'(exists ({" ".join(vars_string_list)})\n {args[0]})'

    def walk_forall(self, expression, args):
        assert len(args) == 1
        vars_string_list = [
            f"{self.get_mangled_name(v)} - {self.get_mangled_name(v.type)}"
            for v in expression.variables()
        ]
        return f'(forall ({" ".join(vars_string_list)})\n {args[0]})'

    def walk_variable_exp(self, expression, args):
        assert len(args) == 0
        return f"{self.get_mangled_name(expression.variable())}"

    def walk_and(self, expression, args):
        assert len(args) > 1
        return f'(and {" ".join(args)})'

    def walk_or(self, expression, args):
        assert len(args) > 1
        return f'(or {" ".join(args)})'

    def walk_not(self, expression, args):
        assert len(args) == 1
        return f"(not {args[0]})"

    def walk_implies(self, expression, args):
        assert len(args) == 2
        return f"(imply {args[0]} {args[1]})"

    def walk_iff(self, expression, args):
        assert len(args) == 2
        return f"(and (imply {args[0]} {args[1]}) (imply {args[1]} {args[0]}) )"

    def walk_fluent_exp(self, expression, args):
        fluent = expression.fluent()
        return f'({self.get_mangled_name(fluent)}{" " if len(args) > 0 else ""}{" ".join(args)})'

    def walk_param_exp(self, expression, args):
        assert len(args) == 0
        p = expression.parameter()
        return f"{self.get_mangled_name(p)}"

    def walk_object_exp(self, expression, args):
        assert len(args) == 0
        o = expression.object()
        return f"{self.get_mangled_name(o)}"

    def walk_bool_constant(self, expression, args):
        raise up.exceptions.UPUnreachableCodeError

    def walk_real_constant(self, expression, args):
        assert len(args) == 0
        frac = expression.constant_value()

        with localcontext() as ctx:
            ctx.prec = self.DECIMAL_PRECISION
            dec = frac.numerator / Decimal(frac.denominator, ctx)

            if Fraction(dec) != frac:
                warn(
                    "The PDDL printer cannot exactly represent the real constant '%s'"
                    % frac
                )
            return str(dec)

    def walk_int_constant(self, expression, args):
        assert len(args) == 0
        return str(expression.constant_value())

    def walk_plus(self, expression, args):
        assert len(args) > 1
        return reduce(lambda x, y: f"(+ {y} {x})", args)

    def walk_minus(self, expression, args):
        assert len(args) == 2
        return f"(- {args[0]} {args[1]})"

    def walk_times(self, expression, args):
        assert len(args) > 1
        return reduce(lambda x, y: f"(* {y} {x})", args)

    def walk_div(self, expression, args):
        assert len(args) == 2
        return f"(/ {args[0]} {args[1]})"

    def walk_le(self, expression, args):
        assert len(args) == 2
        return f"(<= {args[0]} {args[1]})"

    def walk_lt(self, expression, args):
        assert len(args) == 2
        return f"(< {args[0]} {args[1]})"

    def walk_equals(self, expression, args):
        assert len(args) == 2
        return f"(= {args[0]} {args[1]})"


class PDDLWriter:
    """This class can be used to write a Problem in PDDL."""

    def __init__(self, problem: "up.model.Problem", needs_requirements: bool = True):
        self.problem = problem
        self.problem_kind = self.problem.kind
        self.needs_requirements = needs_requirements
        # otn represents the old to new renamings
        self.otn_renamings: Dict[
            Union[
                "up.model.Type",
                "up.model.Action",
                "up.model.Fluent",
                "up.model.Object",
                "up.model.Parameter",
                "up.model.Variable",
            ],
            str,
        ] = {}
        # nto represents the new to old renamings
        self.nto_renamings: Dict[
            str,
            Union[
                "up.model.Type",
                "up.model.Action",
                "up.model.Fluent",
                "up.model.Object",
                "up.model.Parameter",
                "up.model.Variable",
            ],
        ] = {}
        # those 2 maps are "simmetrical", meaning that "(otn[k] == v) implies (nto[v] == k)"
        self.domain_objects: Optional[Dict[_UserType, Set[Object]]] = None

    def _write_domain(self, out: IO[str]):
        if self.problem_kind.has_intermediate_conditions_and_effects():
            raise UPProblemDefinitionError(
                "PDDL2.1 does not support ICE.\nICE are Intermediate Conditions and Effects therefore when an Effect (or Condition) are not at StartTIming(0) or EndTIming(0)."
            )
        if self.problem_kind.has_timed_effect() or self.problem_kind.has_timed_goals():
            raise UPProblemDefinitionError(
                "PDDL2.1 does not support timed effects or timed goals."
            )
        obe = ObjectsExtractor()
        out.write("(define ")
        if self.problem.name is None:
            name = "pddl"
        else:
            name = _get_pddl_name(self.problem)
        out.write(f"(domain {name}-domain)\n")

        if self.needs_requirements:
            out.write(" (:requirements :strips")
            if self.problem_kind.has_flat_typing():
                out.write(" :typing")
            if self.problem_kind.has_negative_conditions():
                out.write(" :negative-preconditions")
            if self.problem_kind.has_disjunctive_conditions():
                out.write(" :disjunctive-preconditions")
            if self.problem_kind.has_equality():
                out.write(" :equality")
            if (
                self.problem_kind.has_continuous_numbers()
                or self.problem_kind.has_discrete_numbers()
            ):
                out.write(" :numeric-fluents")
            if self.problem_kind.has_conditional_effects():
                out.write(" :conditional-effects")
            if self.problem_kind.has_existential_conditions():
                out.write(" :existential-preconditions")
            if self.problem_kind.has_universal_conditions():
                out.write(" :universal-preconditions")
            if (
                self.problem_kind.has_continuous_time()
                or self.problem_kind.has_discrete_time()
            ):
                out.write(" :durative-actions")
            if self.problem_kind.has_duration_inequalities():
                out.write(" :duration-inequalities")
            if (
                self.problem_kind.has_actions_cost()
                or self.problem_kind.has_plan_length()
            ):
                out.write(" :action-costs")
            out.write(")\n")

        if self.problem_kind.has_hierarchical_typing():
            user_types_hierarchy = self.problem.user_types_hierarchy
            out.write(f" (:types\n")
            stack: List["unified_planning.model.Type"] = (
                user_types_hierarchy[None] if None in user_types_hierarchy else []
            )
            out.write(
                f'    {" ".join(self._get_mangled_name(t) for t in stack)} - object\n'
            )
            while stack:
                current_type = stack.pop()
                direct_sons: List["unified_planning.model.Type"] = user_types_hierarchy[
                    current_type
                ]
                if direct_sons:
                    stack.extend(direct_sons)
                    out.write(
                        f'    {" ".join([self._get_mangled_name(t) for t in direct_sons])} - {self._get_mangled_name(current_type)}\n'
                    )
            out.write(" )\n")
        else:
            out.write(
                f' (:types {" ".join([self._get_mangled_name(t) for t in self.problem.user_types])})\n'
                if len(self.problem.user_types) > 0
                else ""
            )

        if self.domain_objects is None:
            # This method populates the self._domain_objects map
            self._populate_domain_objects(obe)
        assert self.domain_objects is not None

        if len(self.domain_objects) > 0:
            out.write(" (:constants")
            for ut, os in self.domain_objects.items():
                if len(os) > 0:
                    out.write(
                        f'\n   {" ".join([self._get_mangled_name(o) for o in os])} - {self._get_mangled_name(ut)}'
                    )
            out.write("\n )\n")

        predicates = []
        functions = []
        for f in self.problem.fluents:
            if f.type.is_bool_type():
                params = []
                i = 0
                for param in f.signature:
                    if param.type.is_user_type():
                        params.append(
                            f" {self._get_mangled_name(param)} - {self._get_mangled_name(param.type)}"
                        )
                        i += 1
                    else:
                        raise UPTypeError("PDDL supports only user type parameters")
                predicates.append(f'({self._get_mangled_name(f)}{"".join(params)})')
            elif f.type.is_int_type() or f.type.is_real_type():
                params = []
                i = 0
                for param in f.signature:
                    if param.type.is_user_type():
                        params.append(
                            f" {self._get_mangled_name(param)} - {self._get_mangled_name(param.type)}"
                        )
                        i += 1
                    else:
                        raise UPTypeError("PDDL supports only user type parameters")
                functions.append(f'({self._get_mangled_name(f)}{"".join(params)})')
            else:
                raise UPTypeError("PDDL supports only boolean and numerical fluents")
        if self.problem.kind.has_actions_cost() or self.problem.kind.has_plan_length():
            functions.append("(total-cost)")
        out.write(
            f' (:predicates {" ".join(predicates)})\n' if len(predicates) > 0 else ""
        )
        out.write(
            f' (:functions {" ".join(functions)})\n' if len(functions) > 0 else ""
        )

        converter = ConverterToPDDLString(self.problem.env, self._get_mangled_name)
        costs = {}
        metrics = self.problem.quality_metrics
        if len(metrics) == 1:
            metric = metrics[0]
            if isinstance(metric, up.model.metrics.MinimizeActionCosts):
                for a in self.problem.actions:
                    cost_exp = metric.get_action_cost(a)
                    costs[a] = cost_exp
                    if cost_exp is not None:
                        _update_domain_objects(self.domain_objects, obe.get(cost_exp))
            elif isinstance(metric, up.model.metrics.MinimizeSequentialPlanLength):
                for a in self.problem.actions:
                    costs[a] = self.problem.env.expression_manager.Int(1)
        elif len(metrics) > 1:
            raise up.exceptions.UPUnsupportedProblemTypeError(
                "Only one metric is supported!"
            )

        for a in self.problem.actions:
            if isinstance(a, up.model.InstantaneousAction):
                out.write(f" (:action {self._get_mangled_name(a)}")
                out.write(f"\n  :parameters (")
                for ap in a.parameters:
                    if ap.type.is_user_type():
                        out.write(
                            f" {self._get_mangled_name(ap)} - {self._get_mangled_name(ap.type)}"
                        )
                    else:
                        raise UPTypeError("PDDL supports only user type parameters")
                out.write(")")
                if len(a.preconditions) > 0:
                    out.write(
                        f'\n  :precondition (and {" ".join([converter.convert(p) for p in a.preconditions])})'
                    )
                if len(a.effects) > 0:
                    out.write("\n  :effect (and")
                    for e in a.effects:
                        if e.is_conditional():
                            out.write(f" (when {converter.convert(e.condition)}")
                        if e.value.is_true():
                            out.write(f" {converter.convert(e.fluent)}")
                        elif e.value.is_false():
                            out.write(f" (not {converter.convert(e.fluent)})")
                        elif e.is_increase():
                            out.write(
                                f" (increase {converter.convert(e.fluent)} {converter.convert(e.value)})"
                            )
                        elif e.is_decrease():
                            out.write(
                                f" (decrease {converter.convert(e.fluent)} {converter.convert(e.value)})"
                            )
                        else:
                            out.write(
                                f" (assign {converter.convert(e.fluent)} {converter.convert(e.value)})"
                            )
                        if e.is_conditional():
                            out.write(f")")

                    if a in costs:
                        out.write(
                            f" (increase (total-cost) {converter.convert(costs[a])})"
                        )
                    out.write(")")
                out.write(")\n")
            elif isinstance(a, DurativeAction):
                out.write(f" (:durative-action {self._get_mangled_name(a)}")
                out.write(f"\n  :parameters (")
                for ap in a.parameters:
                    if ap.type.is_user_type():
                        out.write(
                            f" {self._get_mangled_name(ap)} - {self._get_mangled_name(ap.type)}"
                        )
                    else:
                        raise UPTypeError("PDDL supports only user type parameters")
                out.write(")")
                l, r = a.duration.lower, a.duration.upper
                if l == r:
                    out.write(f"\n  :duration (= ?duration {converter.convert(l)})")
                else:
                    out.write(f"\n  :duration (and ")
                    if a.duration.is_left_open():
                        out.write(f"(> ?duration {converter.convert(l)})")
                    else:
                        out.write(f"(>= ?duration {converter.convert(l)})")
                    if a.duration.is_right_open():
                        out.write(f"(< ?duration {converter.convert(r)})")
                    else:
                        out.write(f"(<= ?duration {converter.convert(r)})")
                    out.write(")")
                if len(a.conditions) > 0:
                    out.write(f"\n  :condition (and ")
                    for interval, cl in a.conditions.items():
                        for c in cl:
                            if interval.lower == interval.upper:
                                if interval.lower.is_from_start():
                                    out.write(f"(at start {converter.convert(c)})")
                                else:
                                    out.write(f"(at end {converter.convert(c)})")
                            else:
                                if not interval.is_left_open():
                                    out.write(f"(at start {converter.convert(c)})")
                                out.write(f"(over all {converter.convert(c)})")
                                if not interval.is_right_open():
                                    out.write(f"(at end {converter.convert(c)})")
                    out.write(")")
                if len(a.effects) > 0:
                    out.write("\n  :effect (and")
                    for t, el in a.effects.items():
                        for e in el:
                            if t.is_from_start():
                                out.write(f" (at start")
                            else:
                                out.write(f" (at end")
                            if e.is_conditional():
                                out.write(f" (when {converter.convert(e.condition)}")
                            if e.value.is_true():
                                out.write(f" {converter.convert(e.fluent)}")
                            elif e.value.is_false():
                                out.write(f" (not {converter.convert(e.fluent)})")
                            elif e.is_increase():
                                out.write(
                                    f" (increase {converter.convert(e.fluent)} {converter.convert(e.value)})"
                                )
                            elif e.is_decrease():
                                out.write(
                                    f" (decrease {converter.convert(e.fluent)} {converter.convert(e.value)})"
                                )
                            else:
                                out.write(
                                    f" (assign {converter.convert(e.fluent)} {converter.convert(e.value)})"
                                )
                            if e.is_conditional():
                                out.write(f")")
                            out.write(")")
                    if a in costs:
                        out.write(
                            f" (at end (increase (total-cost) {converter.convert(costs[a])}))"
                        )
                    out.write(")")
                out.write(")\n")
            else:
                raise NotImplementedError
        out.write(")\n")

    def _write_problem(self, out: IO[str]):
        if self.problem.name is None:
            name = "pddl"
        else:
            name = _get_pddl_name(self.problem)
        out.write(f"(define (problem {name}-problem)\n")
        out.write(f" (:domain {name}-domain)\n")
        if self.domain_objects is None:
            # This method populates the self._domain_objects map
            self._populate_domain_objects(ObjectsExtractor())
        assert self.domain_objects is not None
        if len(self.problem.user_types) > 0:
            out.write(" (:objects")
            for t in self.problem.user_types:
                constants_of_this_type = self.domain_objects.get(
                    cast(_UserType, t), None
                )
                if constants_of_this_type is None:
                    objects = [o for o in self.problem.all_objects if o.type == t]
                else:
                    objects = [
                        o
                        for o in self.problem.all_objects
                        if o.type == t and o not in constants_of_this_type
                    ]
                if len(objects) > 0:
                    out.write(
                        f'\n   {" ".join([self._get_mangled_name(o) for o in objects])} - {self._get_mangled_name(t)}'
                    )
            out.write("\n )\n")
        converter = ConverterToPDDLString(self.problem.env, self._get_mangled_name)
        out.write(" (:init")
        for f, v in self.problem.initial_values.items():
            if v.is_true():
                out.write(f" {converter.convert(f)}")
            elif v.is_false():
                pass
            else:
                out.write(f" (= {converter.convert(f)} {converter.convert(v)})")
        if self.problem.kind.has_actions_cost():
            out.write(f" (= (total-cost) 0)")
        out.write(")\n")
        out.write(
            f' (:goal (and {" ".join([converter.convert(p) for p in self.problem.goals])}))\n'
        )
        metrics = self.problem.quality_metrics
        if len(metrics) == 1:
            metric = metrics[0]
            out.write(" (:metric ")
            if isinstance(metric, up.model.metrics.MinimizeExpressionOnFinalState):
                out.write(f"minimize {converter.convert(metric.expression)}")
            elif isinstance(metric, up.model.metrics.MaximizeExpressionOnFinalState):
                out.write(f"maximize {converter.convert(metric.expression)}")
            elif isinstance(metric, up.model.metrics.MinimizeActionCosts) or isinstance(
                metric, up.model.metrics.MinimizeSequentialPlanLength
            ):
                out.write(f"minimize (total-cost)")
            elif isinstance(metric, up.model.metrics.MinimizeMakespan):
                out.write(f"minimize (total-time)")
            else:
                raise NotImplementedError
            out.write(")\n")
        elif len(metrics) > 1:
            raise up.exceptions.UPUnsupportedProblemTypeError(
                "Only one metric is supported!"
            )
        out.write(")\n")

    def print_domain(self):
        """Prints to std output the PDDL domain."""
        self._write_domain(sys.stdout)

    def print_problem(self):
        """Prints to std output the PDDL problem."""
        self._write_problem(sys.stdout)

    def get_domain(self) -> str:
        """Returns the PDDL domain."""
        out = StringIO()
        self._write_domain(out)
        return out.getvalue()

    def get_problem(self) -> str:
        """Returns the PDDL problem."""
        out = StringIO()
        self._write_problem(out)
        return out.getvalue()

    def write_domain(self, filename: str):
        """Dumps to file the PDDL domain."""
        with open(filename, "w") as f:
            self._write_domain(f)

    def write_problem(self, filename: str):
        """Dumps to file the PDDL problem."""
        with open(filename, "w") as f:
            self._write_problem(f)

    def _get_mangled_name(
        self,
        item: Union[
            "up.model.Type",
            "up.model.Action",
            "up.model.Fluent",
            "up.model.Object",
            "up.model.Parameter",
            "up.model.Variable",
        ],
    ) -> str:
        """This function returns a valid and unique PDDL name."""

        # If we already encountered this item, return it
        if item in self.otn_renamings:
            return self.otn_renamings[item]

        if isinstance(item, up.model.Type):
            assert item.is_user_type()
            original_name = cast(_UserType, item).name
            tmp_name = _get_pddl_name(item)
            # If the problem is hierarchical and the name is object, we want to change it
            if self.problem_kind.has_hierarchical_typing() and tmp_name == "object":
                tmp_name = f"{tmp_name}_"
        else:
            original_name = item.name
            tmp_name = _get_pddl_name(item)
        # if the pddl valid name is the same of the original one and it does not create conflicts,
        # it can be returned
        if tmp_name == original_name and tmp_name not in self.nto_renamings:
            new_name = tmp_name
        else:
            count = 0
            new_name = tmp_name
            while self.problem.has_name(new_name) or new_name in self.nto_renamings:
                new_name = f"{tmp_name}_{count}"
                count += 1
        assert (
            new_name not in self.nto_renamings
            and new_name not in self.otn_renamings.values()
        )
        self.otn_renamings[item] = new_name
        self.nto_renamings[new_name] = item
        return new_name

    def get_item_named(
        self, name: str
    ) -> Union[
        "up.model.Type",
        "up.model.Action",
        "up.model.Fluent",
        "up.model.Object",
        "up.model.Parameter",
        "up.model.Variable",
    ]:
        """
        This method takes a name used in the pddl domain or pddl problem generated by this PDDLWriter and returns the original
        item in the unified_planning problem.
        """
        try:
            return self.nto_renamings[name]
        except KeyError:
            raise UPException(f"The name {name} does not correspond to any item.")

    def get_pddl_name(
        self,
        item: Union[
            "up.model.Type",
            "up.model.Action",
            "up.model.Fluent",
            "up.model.Object",
            "up.model.Parameter",
            "up.model.Variable",
        ],
    ) -> str:
        """
        This method takes an item in the unified_planning problem and returns the chosen name for the same item in the PDDL problem
        or PDDL domain generated by this PDDLWriter.
        """
        try:
            return self.otn_renamings[item]
        except KeyError:
            raise UPException(
                f"The item {item} does not correspond to any item renamed."
            )

    def _populate_domain_objects(self, obe: ObjectsExtractor):
        self.domain_objects = {}
        # Iterate the actions to retrieve domain objects
        for a in self.problem.actions:
            if isinstance(a, up.model.InstantaneousAction):
                for p in a.preconditions:
                    _update_domain_objects(self.domain_objects, obe.get(p))
                for e in a.effects:
                    if e.is_conditional():
                        _update_domain_objects(
                            self.domain_objects, obe.get(e.condition)
                        )
                    _update_domain_objects(self.domain_objects, obe.get(e.fluent))
                    _update_domain_objects(self.domain_objects, obe.get(e.value))
            elif isinstance(a, DurativeAction):
                _update_domain_objects(self.domain_objects, obe.get(a.duration.lower))
                _update_domain_objects(self.domain_objects, obe.get(a.duration.upper))
                for interval, cl in a.conditions.items():
                    for c in cl:
                        _update_domain_objects(self.domain_objects, obe.get(c))
                for t, el in a.effects.items():
                    for e in el:
                        if e.is_conditional():
                            _update_domain_objects(
                                self.domain_objects, obe.get(e.condition)
                            )
                        _update_domain_objects(self.domain_objects, obe.get(e.fluent))
                        _update_domain_objects(self.domain_objects, obe.get(e.value))


def _get_pddl_name(
    item: Union[
        "up.model.Type",
        "up.model.Action",
        "up.model.Fluent",
        "up.model.Object",
        "up.model.Parameter",
        "up.model.Variable",
        "up.model.Problem",
    ]
) -> str:
    """This function returns a pddl name for the chosen item"""
    name = item.name  # type: ignore
    assert name is not None
    name = name.lower()
    regex = re.compile(r"^[a-zA-Z]+.*")
    if (
        re.match(regex, name) is None
    ):  # If the name does not start with an alphabetic char, we make it start with one.
        name = f'{INITIAL_LETTER.get(type(item), "x")}_{name}'

    name = re.sub("[^0-9a-zA-Z_]", "_", name)  # Substitute non-valid elements with "_"
    while (
        name in PDDL_KEYWORDS
    ):  # If the name is in the keywords, apply an underscore at the end until it is not a keyword anymore.
        name = f"{name}_"
    if isinstance(item, up.model.Parameter) or isinstance(item, up.model.Variable):
        name = f"?{name}"
    return name


def _update_domain_objects(
    dict_to_update: Dict[_UserType, Set[Object]], values: Dict[_UserType, Set[Object]]
) -> None:
    """Small utility method that updated a UserType -> Set[Object] dict with another dict of the same type."""
    for ut, os in values.items():
        os_to_update = dict_to_update.setdefault(ut, set())
        os_to_update |= os



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unified_planning as up
from unified_planning.exceptions import UPProblemDefinitionError, UPValueError
from typing import Iterator, List


class ActionsSetMixin:
    """
    This class is a mixin that contains a set of actions with some related methods.

    NOTE: when this mixin is used in combination with other mixins that share some
    of the attributes (e.g. env, add_user_type_method, has_name_method), it is required
    to pass the very same arguments to the mixins constructors.
    """

    def __init__(self, env, add_user_type_method, has_name_method):
        self._env = env
        self._add_user_type_method = add_user_type_method
        self._has_name_method = has_name_method
        self._actions: List["up.model.action.Action"] = []

    @property
    def env(self) -> "up.environment.Environment":
        """Returns the problem environment."""
        return self._env

    @property
    def actions(self) -> List["up.model.action.Action"]:
        """Returns the list of the actions in the problem."""
        return self._actions

    def clear_actions(self):
        """Removes all the problem actions."""
        self._actions = []

    @property
    def instantaneous_actions(self) -> Iterator["up.model.action.InstantaneousAction"]:
        """Returs all the instantaneous actions of the problem.

        IMPORTANT NOTE: this property does some computation, so it should be called as
        seldom as possible."""
        for a in self._actions:
            if isinstance(a, up.model.action.InstantaneousAction):
                yield a

    @property
    def durative_actions(self) -> Iterator["up.model.action.DurativeAction"]:
        """Returs all the durative actions of the problem.

        IMPORTANT NOTE: this property does some computation, so it should be called as
        seldom as possible."""
        for a in self._actions:
            if isinstance(a, up.model.action.DurativeAction):
                yield a

    @property
    def conditional_actions(self) -> List["up.model.action.Action"]:
        """Returns the conditional actions.

        IMPORTANT NOTE: this property does some computation, so it should be called as
        seldom as possible."""
        return [a for a in self._actions if a.is_conditional()]

    @property
    def unconditional_actions(self) -> List["up.model.action.Action"]:
        """Returns the conditional actions.

        IMPORTANT NOTE: this property does some computation, so it should be called as
        seldom as possible."""
        return [a for a in self._actions if not a.is_conditional()]

    def action(self, name: str) -> "up.model.action.Action":
        """Returns the action with the given name."""
        for a in self._actions:
            if a.name == name:
                return a
        raise UPValueError(f"Action of name: {name} is not defined!")

    def has_action(self, name: str) -> bool:
        """Returns True if the problem has the action with the given name ."""
        for a in self._actions:
            if a.name == name:
                return True
        return False

    def add_action(self, action: "up.model.action.Action"):
        """Adds the given action."""
        assert (
            action.env == self._env
        ), "Action does not have the same environment of the problem"
        if self._has_name_method(action.name):
            raise UPProblemDefinitionError("Name " + action.name + " already defined!")
        self._actions.append(action)
        for param in action.parameters:
            if param.type.is_user_type():
                self._add_user_type_method(param.type)

    def add_actions(self, actions: List["up.model.action.Action"]):
        """Adds the given actions."""
        for action in actions:
            self.add_action(action)



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unified_planning as up
import unified_planning.environment
import unified_planning.engines as engines
import unified_planning.engines.compilers
from unified_planning.plans import ActionInstance
from unified_planning.engines.mixins.compiler import CompilationKind, CompilerMixin
from unified_planning.engines.results import CompilerResult
from unified_planning.model import Problem, ProblemKind, Action, Type, Expression, FNode
from unified_planning.model.types import domain_size, domain_item
from unified_planning.model.walkers import Substituter, Simplifier
from unified_planning.engines.compilers.utils import (
    lift_action_instance,
    create_action_with_given_subs,
)
from typing import Dict, List, Optional, Tuple, Iterator
from itertools import product
from functools import partial


class Grounder(engines.engine.Engine, CompilerMixin):
    """Performs action grounding."""

    def __init__(
        self,
        grounding_actions_map: Optional[Dict[Action, List[Tuple[FNode, ...]]]] = None,
    ):
        self._grounding_actions_map = grounding_actions_map

    @property
    def name(self):
        return "grounder"

    @staticmethod
    def supported_kind() -> ProblemKind:
        supported_kind = ProblemKind()
        supported_kind.set_problem_class("ACTION_BASED")
        supported_kind.set_typing("FLAT_TYPING")
        supported_kind.set_typing("HIERARCHICAL_TYPING")
        supported_kind.set_numbers("CONTINUOUS_NUMBERS")
        supported_kind.set_numbers("DISCRETE_NUMBERS")
        supported_kind.set_fluents_type("NUMERIC_FLUENTS")
        supported_kind.set_fluents_type("OBJECT_FLUENTS")
        supported_kind.set_conditions_kind("NEGATIVE_CONDITIONS")
        supported_kind.set_conditions_kind("DISJUNCTIVE_CONDITIONS")
        supported_kind.set_conditions_kind("EQUALITY")
        supported_kind.set_conditions_kind("EXISTENTIAL_CONDITIONS")
        supported_kind.set_conditions_kind("UNIVERSAL_CONDITIONS")
        supported_kind.set_effects_kind("CONDITIONAL_EFFECTS")
        supported_kind.set_effects_kind("INCREASE_EFFECTS")
        supported_kind.set_effects_kind("DECREASE_EFFECTS")
        supported_kind.set_time("CONTINUOUS_TIME")
        supported_kind.set_time("DISCRETE_TIME")
        supported_kind.set_time("INTERMEDIATE_CONDITIONS_AND_EFFECTS")
        supported_kind.set_time("TIMED_EFFECT")
        supported_kind.set_time("TIMED_GOALS")
        supported_kind.set_time("DURATION_INEQUALITIES")
        supported_kind.set_simulated_entities("SIMULATED_EFFECTS")
        return supported_kind

    @staticmethod
    def supports(problem_kind):
        return problem_kind <= Grounder.supported_kind()

    @staticmethod
    def supports_compilation(compilation_kind: CompilationKind) -> bool:
        return compilation_kind == CompilationKind.GROUNDING

    def _compile(
        self,
        problem: "up.model.AbstractProblem",
        compilation_kind: "up.engines.CompilationKind",
    ) -> CompilerResult:
        assert isinstance(problem, Problem)

        env = problem.env
        substituter = Substituter(env)
        simplifier = Simplifier(env, problem)
        trace_back_map: Dict[Action, Tuple[Action, List[FNode]]] = {}

        new_problem = problem.clone()
        new_problem.name = f"{self.name}_{problem.name}"
        new_problem.clear_actions()
        for old_action in problem.actions:
            # contains the type of every parameter of the action
            type_list: List[Type] = [param.type for param in old_action.parameters]
            # if the action does not have parameters, it does not need to be grounded.
            if len(type_list) == 0:
                if (
                    self._grounding_actions_map is None
                    or self._grounding_actions_map.get(old_action, None) is not None
                ):
                    new_action = old_action.clone()
                    new_problem.add_action(new_action)
                    trace_back_map[new_action] = (old_action, [])
                continue
            grounded_params_list: Optional[Iterator[Tuple[FNode, ...]]] = None
            if self._grounding_actions_map is None:
                # a list containing the list of object in the problem of the given type.
                # So, if the problem has 2 Locations l1 and l2, and 2 Robots r1 and r2, and
                # the action move_to takes as parameters a Robot and a Location,
                # the variable state at this point will be the following:
                # type_list = [Robot, Location]
                # objects_list = [[r1, r2], [l1, l2]]
                # the product of *objects_list will be:
                # [(r1, l1), (r1, l2), (r2, l1), (r2,l2)]
                ground_size = 1
                domain_sizes = []
                for t in type_list:
                    ds = domain_size(new_problem, t)
                    domain_sizes.append(ds)
                    ground_size *= ds
                items_list: List[List[FNode]] = []
                for size, type in zip(domain_sizes, type_list):
                    items_list.append(
                        [domain_item(new_problem, type, j) for j in range(size)]
                    )
                grounded_params_list = product(*items_list)
            else:
                # The grounding_actions_map is not None, therefore it must be used to ground
                grounded_params_list = iter(self._grounding_actions_map[old_action])
            assert grounded_params_list is not None
            for grounded_params in grounded_params_list:
                subs: Dict[Expression, Expression] = dict(
                    zip(old_action.parameters, list(grounded_params))
                )
                new_action = create_action_with_given_subs(
                    problem, old_action, simplifier, substituter, subs
                )
                # when the action is None it means it is not feasible,
                # it's conditions are in contraddiction within one another.
                if new_action is not None:
                    trace_back_map[new_action] = (
                        old_action,
                        env.expression_manager.auto_promote(subs.values()),
                    )
                    new_problem.add_action(new_action)

        return CompilerResult(
            new_problem, partial(lift_action_instance, map=trace_back_map), self.name
        )



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import networkx as nx
import unified_planning as up
import unified_planning.plans as plans
import unified_planning.model.walkers as walkers
from unified_planning.environment import Environment
from unified_planning.exceptions import UPUsageError
from unified_planning.model import FNode, InstantaneousAction, Expression
from typing import Callable, Dict, Optional, Set, List, cast


class SequentialPlan(plans.plan.Plan):
    """Represents a sequential plan."""

    def __init__(
        self,
        actions: List["plans.plan.ActionInstance"],
        environment: Optional["Environment"] = None,
    ):
        # if we have a specific env or we don't have any actions
        if environment is not None or not actions:
            plans.plan.Plan.__init__(
                self, plans.plan.PlanKind.SEQUENTIAL_PLAN, environment
            )
        # If we don't have a specific env and have at least 1 action, use the env of the first action
        else:
            assert len(actions) > 0
            plans.plan.Plan.__init__(
                self, plans.plan.PlanKind.SEQUENTIAL_PLAN, actions[0].action.env
            )
        for (
            ai
        ) in actions:  # check that given env and the env in the actions is the same
            if ai.action.env != self._environment:
                raise UPUsageError(
                    "The environment given to the plan is not the same of the actions in the plan."
                )
        self._actions = actions

    def __repr__(self) -> str:
        return str(self._actions)

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, SequentialPlan) and len(self._actions) == len(oth._actions):
            for ai, oth_ai in zip(self._actions, oth._actions):
                if not ai.is_semantically_equivalent(oth_ai):
                    return False
            return True
        else:
            return False

    def __hash__(self) -> int:
        count: int = 0
        for i, ai in enumerate(self._actions):
            count += i + hash(ai.action) + hash(ai.actual_parameters)
        return count

    def __contains__(self, item: object) -> bool:
        if isinstance(item, plans.plan.ActionInstance):
            return any(item.is_semantically_equivalent(a) for a in self._actions)
        else:
            return False

    @property
    def actions(self) -> List["plans.plan.ActionInstance"]:
        """Returns the sequence of action instances."""
        return self._actions

    def replace_action_instances(
        self,
        replace_function: Callable[
            ["plans.plan.ActionInstance"], "plans.plan.ActionInstance"
        ],
    ) -> "up.plans.plan.Plan":
        new_ai = [replace_function(ai) for ai in self._actions]
        new_env = self._environment
        if len(new_ai) > 0:
            new_env = new_ai[0].action.env
        return SequentialPlan(new_ai, new_env)

    def to_partial_order_plan(
        self, problem: "up.model.mixins.ObjectsSetMixin"
    ) -> "up.plans.partial_order_plan.PartialOrderPlan":
        """Returns the PartialOrderPlan version of this SequentialPlan.

        This is done by keeping the ordering constraints, given by the SequentialPlan, between 2 ActionInstances
        that satisfy one of these conditions:
        + at least one of the 2 ActionInstances writes on a grounded fluent (writes means that one of his effects
                assign a value to said fluent)
            - AND the other ActionInstance reads or writes on the same grounded fluent (reads means that one of his preconditions
                or one of his condition in a conditional effect depends on said fluent)."""
        subs = walkers.Substituter(self._environment)
        simp = self._environment.simplifier
        eqr = walkers.ExpressionQuantifiersRemover(self._environment)
        fve = self._environment.free_vars_extractor
        # last_modifier is the mapping from a grounded fluent to the last action instance that assigned a value to
        # that fluent
        last_modifier: Dict[FNode, "plans.plan.ActionInstance"] = {}
        # all_required is the mapping from a grounded fluent to all the action instances that read the value of that
        # fluent in their preconditions (or in the condition of their conditional effects)
        all_required: Dict[FNode, List["plans.plan.ActionInstance"]] = {}
        # graph stores the information gathered through the process
        graph = nx.DiGraph()
        for action_instance in self.actions:
            graph.add_node(action_instance)
            assert isinstance(action_instance.action, InstantaneousAction)
            inst_action = cast(InstantaneousAction, action_instance.action)

            # required_fluents contains all the grounded fluents that this action_instance "reads"
            required_fluents: Set[FNode] = set()
            # same of required_fluents, but the fluents are lifted
            lifted_required_fluents: Set[FNode] = set()
            # add free vars of preconditions
            for prec in inst_action.preconditions:
                lifted_required_fluents |= fve.get(
                    eqr.remove_quantifiers(prec, problem)
                )
            # add in the required_fluents all the free fluents this action instance deals with
            for eff in inst_action.effects:
                lifted_required_fluents |= fve.get(
                    eqr.remove_quantifiers(eff.condition, problem)
                )
                lifted_required_fluents |= fve.get(
                    eqr.remove_quantifiers(eff.fluent, problem)
                )
                lifted_required_fluents |= fve.get(
                    eqr.remove_quantifiers(eff.value, problem)
                )

            assignments: Dict[Expression, Expression] = dict(
                zip(inst_action.parameters, action_instance.actual_parameters)
            )
            for lifted_fluent in lifted_required_fluents:
                assert lifted_fluent.is_fluent_exp()
                for (
                    arg
                ) in lifted_fluent.args:  # check that we don't have "nested" fluents
                    if len(fve.get(eqr.remove_quantifiers(arg, problem))) != 0:
                        raise UPUsageError(
                            f"The partial deordering of a Sequential Plan does not allow the use of fluents inside the parameter of fluents!\nThe fluent: {lifted_fluent} does violates this contraint."
                        )
                required_fluents.add(
                    simp.simplify(subs.substitute(lifted_fluent, assignments))
                )

            # for every required fluent, add this action instance to the list of action instances that requires this fluent
            # and order the current action instance after the last modifier of the fluent
            for required_fluent in required_fluents:
                action_instance_list = all_required.setdefault(required_fluent, [])
                action_instance_list.append(action_instance)
                required_fluent_last_modifier = last_modifier.get(required_fluent, None)
                if required_fluent_last_modifier is not None:
                    assert (
                        required_fluent_last_modifier != action_instance
                    )  # sanity check
                    graph.add_edge(required_fluent_last_modifier, action_instance)

            # for every effect, set current action instance as the last modifier and the current action instance is ordered
            # after every action instance that requires a fluent the current action instance modifies
            for eff in inst_action.effects:
                assert eff.fluent.is_fluent_exp()
                grounded_fluent = simp.simplify(
                    subs.substitute(eff.fluent, assignments)
                )
                last_modifier[grounded_fluent] = action_instance
                dependent_action_instance_list = all_required.setdefault(
                    grounded_fluent, []
                )
                for dependent_action_instance in dependent_action_instance_list:
                    if dependent_action_instance != action_instance:
                        graph.add_edge(dependent_action_instance, action_instance)

        assert nx.is_directed_acyclic_graph(graph)
        # remove redundant edges and return the up.plans.partial_order_plan.PartialOrderPlan structure.
        return up.plans.partial_order_plan.PartialOrderPlan(
            {}, self._environment, nx.transitive_reduction(graph)
        )



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unified_planning
from unified_planning.shortcuts import *
from collections import namedtuple

Example = namedtuple("Example", ["problem", "plan"])


def get_example_problems():
    problems = {}

    # robot_real_constants
    # this version of the problem robot has reals instead of integers as constants
    Location = UserType("Location")
    robot_at = Fluent("robot_at", BoolType(), position=Location)
    battery_charge = Fluent("battery_charge", RealType(0, 100))
    move = InstantaneousAction("move", l_from=Location, l_to=Location)
    l_from = move.parameter("l_from")
    l_to = move.parameter("l_to")
    move.add_precondition(GE(battery_charge, 10.0))
    move.add_precondition(Not(Equals(l_from, l_to)))
    move.add_precondition(robot_at(l_from))
    move.add_precondition(Not(robot_at(l_to)))
    move.add_effect(robot_at(l_from), False)
    move.add_effect(robot_at(l_to), True)
    move.add_effect(battery_charge, Minus(battery_charge, 10.0))
    l1 = Object("l1", Location)
    l2 = Object("l2", Location)
    problem = Problem("robot_real_constants")
    problem.add_fluent(robot_at)
    problem.add_fluent(battery_charge)
    problem.add_action(move)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.set_initial_value(robot_at(l1), True)
    problem.set_initial_value(robot_at(l2), False)
    problem.set_initial_value(battery_charge, 100.0)
    problem.add_goal(robot_at(l2))
    plan = unified_planning.plans.SequentialPlan(
        [unified_planning.plans.ActionInstance(move, (ObjectExp(l1), ObjectExp(l2)))]
    )
    robot = Example(problem=problem, plan=plan)
    problems["robot_real_constants"] = robot

    # robot_int_battery
    # this version of the problem robot has the battery charge fluent represented as an int instead of a real
    Location = UserType("Location")
    robot_at = Fluent("robot_at", BoolType(), position=Location)
    battery_charge = Fluent("battery_charge", IntType(0, 100))
    move = InstantaneousAction("move", l_from=Location, l_to=Location)
    l_from = move.parameter("l_from")
    l_to = move.parameter("l_to")
    move.add_precondition(GE(battery_charge, 10))
    move.add_precondition(Not(Equals(l_from, l_to)))
    move.add_precondition(robot_at(l_from))
    move.add_precondition(Not(robot_at(l_to)))
    move.add_effect(robot_at(l_from), False)
    move.add_effect(robot_at(l_to), True)
    move.add_effect(battery_charge, Minus(battery_charge, 10))
    l1 = Object("l1", Location)
    l2 = Object("l2", Location)
    problem = Problem("robot_int_battery")
    problem.add_fluent(robot_at)
    problem.add_fluent(battery_charge)
    problem.add_action(move)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.set_initial_value(robot_at(l1), True)
    problem.set_initial_value(robot_at(l2), False)
    problem.set_initial_value(battery_charge, 100)
    problem.add_goal(robot_at(l2))
    plan = unified_planning.plans.SequentialPlan(
        [unified_planning.plans.ActionInstance(move, (ObjectExp(l1), ObjectExp(l2)))]
    )
    robot = Example(problem=problem, plan=plan)
    problems["robot_int_battery"] = robot

    # robot fluent of user_type with int ID
    Int_t = IntType(0, 1)
    Location = UserType("Location")
    is_at = Fluent("is_at", Location, id=Int_t)
    move = InstantaneousAction("move", robot=Int_t, l_from=Location, l_to=Location)
    robot = move.parameter("robot")
    l_from = move.parameter("l_from")
    l_to = move.parameter("l_to")
    move.add_precondition(Equals(is_at(robot), l_from))
    move.add_precondition(Not(Equals(is_at(robot), l_to)))
    move.add_effect(is_at(robot), l_to)
    l1 = Object("l1", Location)
    l2 = Object("l2", Location)
    problem = Problem("robot_fluent_of_user_type_with_int_id")
    problem.add_fluent(is_at)
    problem.add_action(move)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.set_initial_value(is_at(Int(0)), l1)
    problem.set_initial_value(is_at(1), l1)
    plan = unified_planning.plans.SequentialPlan(
        [
            unified_planning.plans.ActionInstance(
                move, (Int(0), ObjectExp(l1), ObjectExp(l2))
            ),
            unified_planning.plans.ActionInstance(
                move, (Int(1), ObjectExp(l1), ObjectExp(l2))
            ),
        ]
    )
    robot_fluent_of_user_type_with_int_id = Example(problem=problem, plan=plan)
    problems[
        "robot_fluent_of_user_type_with_int_id"
    ] = robot_fluent_of_user_type_with_int_id

    # robot locations connected without battery
    Location = UserType("Location")
    Robot = UserType("Robot")
    is_at = Fluent("is_at", BoolType(), position=Location, robot=Robot)
    is_connected = Fluent(
        "is_connected", BoolType(), location_1=Location, location_2=Location
    )
    move = InstantaneousAction("move", robot=Robot, l_from=Location, l_to=Location)
    robot = move.parameter("robot")
    l_from = move.parameter("l_from")
    l_to = move.parameter("l_to")
    move.add_precondition(Not(Equals(l_from, l_to)))
    move.add_precondition(is_at(l_from, robot))
    move.add_precondition(Not(is_at(l_to, robot)))
    move.add_precondition(Or(is_connected(l_from, l_to), is_connected(l_to, l_from)))
    move.add_effect(is_at(l_from, robot), False)
    move.add_effect(is_at(l_to, robot), True)
    move_2 = InstantaneousAction("move_2", robot=Robot, l_from=Location, l_to=Location)
    robot = move_2.parameter("robot")
    l_from = move_2.parameter("l_from")
    l_to = move_2.parameter("l_to")
    move_2.add_precondition(Not(Equals(l_from, l_to)))
    move_2.add_precondition(is_at(l_from, robot))
    move_2.add_precondition(Not(is_at(l_to, robot)))
    mid_location = Variable("mid_loc", Location)
    # (E (location mid_location)
    # !((mid_location == l_from) || (mid_location == l_to)) && (is_connected(l_from, mid_location) || is_connected(mid_location, l_from)) &&
    # && (is_connected(l_to, mid_location) || is_connected(mid_location, l_to)))
    move_2.add_precondition(
        Exists(
            And(
                Not(Or(Equals(mid_location, l_from), Equals(mid_location, l_to))),
                Or(
                    is_connected(l_from, mid_location),
                    is_connected(mid_location, l_from),
                ),
                Or(is_connected(l_to, mid_location), is_connected(mid_location, l_to)),
            ),
            mid_location,
        )
    )
    move_2.add_effect(is_at(l_from, robot), False)
    move_2.add_effect(is_at(l_to, robot), True)
    l1 = Object("l1", Location)
    l2 = Object("l2", Location)
    l3 = Object("l3", Location)
    l4 = Object("l4", Location)
    l5 = Object("l5", Location)
    r1 = Object("r1", Robot)
    problem = Problem("robot_locations_connected_without_battery")
    problem.add_fluent(is_at, default_initial_value=False)
    problem.add_fluent(is_connected, default_initial_value=False)
    problem.add_action(move)
    problem.add_action(move_2)
    problem.add_object(r1)
    problem.add_object(l1)
    problem.add_object(l2)
    problem.add_object(l3)
    problem.add_object(l4)
    problem.add_object(l5)
    problem.set_initial_value(is_at(l1, r1), True)
    problem.set_initial_value(is_connected(l1, l2), True)
    problem.set_initial_value(is_connected(l2, l3), True)
    problem.set_initial_value(is_connected(l3, l4), True)
    problem.set_initial_value(is_connected(l4, l5), True)
    problem.add_goal(is_at(l5, r1))
    plan = unified_planning.plans.SequentialPlan(
        [
            unified_planning.plans.ActionInstance(
                move_2, (ObjectExp(r1), ObjectExp(l1), ObjectExp(l3))
            ),
            unified_planning.plans.ActionInstance(
                move_2, (ObjectExp(r1), ObjectExp(l3), ObjectExp(l5))
            ),
        ]
    )
    robot_locations_connected_without_battery = Example(problem=problem, plan=plan)
    problems[
        "robot_locations_connected_without_battery"
    ] = robot_locations_connected_without_battery

    # hierarchical blocks world exists
    Entity = UserType("Entity", None)  # None can be avoided
    Location = UserType("Location", Entity)
    Unmovable = UserType("Unmovable", Location)
    TableSpace = UserType("TableSpace", Unmovable)
    Movable = UserType("Movable", Location)
    Block = UserType("Block", Movable)
    clear = Fluent("clear", BoolType(), space=Location)
    on = Fluent("on", BoolType(), object=Movable, space=Location)

    move = InstantaneousAction("move", item=Movable, l_from=Location, l_to=Location)
    item = move.parameter("item")
    l_from = move.parameter("l_from")
    l_to = move.parameter("l_to")
    move.add_precondition(clear(item))
    move.add_precondition(clear(l_to))
    move.add_precondition(on(item, l_from))
    move.add_effect(clear(l_from), True)
    move.add_effect(on(item, l_from), False)
    move.add_effect(clear(l_to), False)
    move.add_effect(on(item, l_to), True)

    problem = Problem("hierarchical_blocks_world_exists")
    problem.add_fluent(clear, default_initial_value=False)
    problem.add_fluent(on, default_initial_value=False)
    problem.add_action(move)
    ts_1 = Object("ts_1", TableSpace)
    ts_2 = Object("ts_2", TableSpace)
    ts_3 = Object("ts_3", TableSpace)
    problem.add_objects([ts_1, ts_2, ts_3])
    block_1 = Object("block_1", Block)
    block_2 = Object("block_2", Block)
    block_3 = Object("block_3", Block)
    problem.add_objects([block_1, block_2, block_3])

    # The blocks are all on ts_1, in order block_3 under block_1 under block_2
    problem.set_initial_value(clear(ts_2), True)
    problem.set_initial_value(clear(ts_3), True)
    problem.set_initial_value(clear(block_2), True)
    problem.set_initial_value(on(block_3, ts_1), True)
    problem.set_initial_value(on(block_1, block_3), True)
    problem.set_initial_value(on(block_2, block_1), True)

    # We want them on ts_3 in order block_3 on block_2 on block_1
    problem.add_goal(on(block_1, ts_3))
    m_var = Variable("m_var", Movable)
    problem.add_goal(Exists(on(m_var, block_1), m_var))
    problem.add_goal(on(block_3, block_2))

    plan = unified_planning.plans.SequentialPlan(
        [
            unified_planning.plans.ActionInstance(
                move, (ObjectExp(block_2), ObjectExp(block_1), ObjectExp(ts_2))
            ),
            unified_planning.plans.ActionInstance(
                move, (ObjectExp(block_1), ObjectExp(block_3), ObjectExp(ts_3))
            ),
            unified_planning.plans.ActionInstance(
                move, (ObjectExp(block_2), ObjectExp(ts_2), ObjectExp(block_1))
            ),
            unified_planning.plans.ActionInstance(
                move, (ObjectExp(block_3), ObjectExp(ts_1), ObjectExp(block_2))
            ),
        ]
    )
    hierarchical_blocks_world_exists = Example(problem=problem, plan=plan)
    problems["hierarchical_blocks_world_exists"] = hierarchical_blocks_world_exists

    # hierarchical blocks world object as root
    object = UserType("object")
    Entity = UserType("Entity", object)
    Location = UserType("Location", Entity)
    Unmovable = UserType("Unmovable", Location)
    TableSpace = UserType("TableSpace", Unmovable)
    Movable = UserType("Movable", Location)
    Block = UserType("Block", Movable)
    clear = Fluent("clear", BoolType(), space=Location)
    on = Fluent("on", BoolType(), object=Movable, space=Location)

    move = InstantaneousAction("move", item=Movable, l_from=Location, l_to=Location)
    item = move.parameter("item")
    l_from = move.parameter("l_from")
    l_to = move.parameter("l_to")
    move.add_precondition(clear(item))
    move.add_precondition(clear(l_to))
    move.add_precondition(on(item, l_from))
    move.add_effect(clear(l_from), True)
    move.add_effect(on(item, l_from), False)
    move.add_effect(clear(l_to), False)
    move.add_effect(on(item, l_to), True)

    problem = Problem("hierarchical_blocks_world_object_as_root")
    problem.add_fluent(clear, default_initial_value=False)
    problem.add_fluent(on, default_initial_value=False)
    problem.add_action(move)
    ts_1 = Object("ts_1", TableSpace)
    ts_2 = Object("ts_2", TableSpace)
    ts_3 = Object("ts_3", TableSpace)
    problem.add_objects([ts_1, ts_2, ts_3])
    block_1 = Object("block_1", Block)
    block_2 = Object("block_2", Block)
    block_3 = Object("block_3", Block)
    problem.add_objects([block_1, block_2, block_3])

    # The blocks are all on ts_1, in order block_3 under block_1 under block_2
    problem.set_initial_value(clear(ts_2), True)
    problem.set_initial_value(clear(ts_3), True)
    problem.set_initial_value(clear(block_2), True)
    problem.set_initial_value(on(block_3, ts_1), True)
    problem.set_initial_value(on(block_1, block_3), True)
    problem.set_initial_value(on(block_2, block_1), True)

    # We want them on ts_3 in order block_3 on block_2 on block_1
    problem.add_goal(on(block_1, ts_3))
    problem.add_goal(on(block_2, block_1))
    problem.add_goal(on(block_3, block_2))

    plan = unified_planning.plans.SequentialPlan(
        [
            unified_planning.plans.ActionInstance(
                move, (ObjectExp(block_2), ObjectExp(block_1), ObjectExp(ts_2))
            ),
            unified_planning.plans.ActionInstance(
                move, (ObjectExp(block_1), ObjectExp(block_3), ObjectExp(ts_3))
            ),
            unified_planning.plans.ActionInstance(
                move, (ObjectExp(block_2), ObjectExp(ts_2), ObjectExp(block_1))
            ),
            unified_planning.plans.ActionInstance(
                move, (ObjectExp(block_3), ObjectExp(ts_1), ObjectExp(block_2))
            ),
        ]
    )
    hierarchical_blocks_world_object_as_root = Example(problem=problem, plan=plan)
    problems[
        "hierarchical_blocks_world_object_as_root"
    ] = hierarchical_blocks_world_object_as_root

    # hierarchical blocks world with object
    Entity = UserType("Entity", None)  # None can be avoided
    object = UserType("object", Entity)
    Unmovable = UserType("Unmovable", object)
    TableSpace = UserType("TableSpace", Unmovable)
    Movable = UserType("Movable", object)
    Block = UserType("Block", Movable)
    clear = Fluent("clear", BoolType(), space=object)
    on = Fluent("on", BoolType(), object=Movable, space=object)

    move = InstantaneousAction("move", item=Movable, l_from=object, l_to=object)
    item = move.parameter("item")
    l_from = move.parameter("l_from")
    l_to = move.parameter("l_to")
    move.add_precondition(clear(item))
    move.add_precondition(clear(l_to))
    move.add_precondition(on(item, l_from))
    move.add_effect(clear(l_from), True)
    move.add_effect(on(item, l_from), False)
    move.add_effect(clear(l_to), False)
    move.add_effect(on(item, l_to), True)

    problem = Problem("hierarchical_blocks_world_with_object")
    problem.add_fluent(clear, default_initial_value=False)
    problem.add_fluent(on, default_initial_value=False)
    problem.add_action(move)
    ts_1 = Object("ts_1", TableSpace)
    ts_2 = Object("ts_2", TableSpace)
    ts_3 = Object("ts_3", TableSpace)
    problem.add_objects([ts_1, ts_2, ts_3])
    block_1 = Object("block_1", Block)
    block_2 = Object("block_2", Block)
    block_3 = Object("block_3", Block)
    problem.add_objects([block_1, block_2, block_3])

    # The blocks are all on ts_1, in order block_3 under block_1 under block_2
    problem.set_initial_value(clear(ts_2), True)
    problem.set_initial_value(clear(ts_3), True)
    problem.set_initial_value(clear(block_2), True)
    problem.set_initial_value(on(block_3, ts_1), True)
    problem.set_initial_value(on(block_1, block_3), True)
    problem.set_initial_value(on(block_2, block_1), True)

    # We want them on ts_3 in order block_3 on block_2 on block_1
    problem.add_goal(on(block_1, ts_3))
    problem.add_goal(on(block_2, block_1))
    problem.add_goal(on(block_3, block_2))

    plan = unified_planning.plans.SequentialPlan(
        [
            unified_planning.plans.ActionInstance(
                move, (ObjectExp(block_2), ObjectExp(block_1), ObjectExp(ts_2))
            ),
            unified_planning.plans.ActionInstance(
                move, (ObjectExp(block_1), ObjectExp(block_3), ObjectExp(ts_3))
            ),
            unified_planning.plans.ActionInstance(
                move, (ObjectExp(block_2), ObjectExp(ts_2), ObjectExp(block_1))
            ),
            unified_planning.plans.ActionInstance(
                move, (ObjectExp(block_3), ObjectExp(ts_1), ObjectExp(block_2))
            ),
        ]
    )
    hierarchical_blocks_world_with_object = Example(problem=problem, plan=plan)
    problems[
        "hierarchical_blocks_world_with_object"
    ] = hierarchical_blocks_world_with_object

    return problems



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unified_planning as up
from unified_planning.shortcuts import *
from unified_planning.model.problem_kind import full_classical_kind, full_numeric_kind
from unified_planning.test import TestCase, main, skipIfEngineNotAvailable
from unified_planning.test.examples import get_example_problems


class TestPartialOrderPlan(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()

    @skipIfEngineNotAvailable("sequential_plan_validator")
    def test_all(self):
        with PlanValidator(name="sequential_plan_validator") as validator:
            assert validator is not None
            for problem, plan in self.problems.values():
                if validator.supports(problem.kind):
                    self.assertTrue(isinstance(plan, up.plans.SequentialPlan))
                    pop_plan = plan.to_partial_order_plan(problem)
                    for sorted_plan in pop_plan.all_sequential_plans():
                        validation_result = validator.validate(problem, sorted_plan)
                        self.assertEqual(
                            up.engines.ValidationResultStatus.VALID,
                            validation_result.status,
                        )



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from typing import Dict, List, Optional
from itertools import product

import unified_planning as up
import unified_planning.environment
from unified_planning.exceptions import UPProblemDefinitionError
from unified_planning.model.fnode import FNode
from unified_planning.model.expression import Expression
from unified_planning.model.walkers.dag import DagWalker
from unified_planning.model.walkers.simplifier import Simplifier


class QuantifierSimplifier(Simplifier):
    """Same to the unified_planning.Simplifier, but does not expand quantifiers and solves them locally."""

    def __init__(
        self,
        env: "unified_planning.environment.Environment",
        problem: "up.model.problem.Problem",
    ):
        DagWalker.__init__(self, True)
        self._env = env
        self.manager = env.expression_manager
        self._problem = problem
        self._assignments: Optional[Dict["Expression", "Expression"]] = None
        self._variable_assignments: Optional[Dict["Expression", "Expression"]] = None

    def qsimplify(
        self,
        expression: "FNode",
        assignments: Dict["Expression", "Expression"],
        variable_assignments: Dict["Expression", "Expression"],
    ):
        assert self._problem is not None
        assert self._assignments is None
        assert self._variable_assignments is None
        self._assignments = assignments
        self._variable_assignments = variable_assignments
        r = self.walk(expression)
        self._assignments = None
        self._variable_assignments = None
        return r

    def _push_with_children_to_stack(self, expression: "FNode", **kwargs):
        """Add children to the stack."""
        if expression.is_forall() or expression.is_exists():
            self.stack.append((True, expression))
        else:
            super(Simplifier, self)._push_with_children_to_stack(expression, **kwargs)

    def _compute_node_result(self, expression: "FNode", **kwargs):
        """Apply function to the node and memoize the result.
        Note: This function assumes that the results for the children
              are already available.
        """
        key = self._get_key(expression, **kwargs)
        if key not in self.memoization:
            try:
                f = self.functions[expression.node_type]
            except KeyError:
                f = self.walk_error

            if not (expression.is_forall() or expression.is_exists()):
                args = [
                    self.memoization[self._get_key(s, **kwargs)]
                    for s in self._get_children(expression)
                ]
                self.memoization[key] = f(expression, args=args, **kwargs)
            else:
                self.memoization[key] = f(expression, args=expression.args, **kwargs)
        else:
            pass

    def _deep_subs_simplify(
        self,
        expression: "FNode",
        variables_assignments: Dict["Expression", "Expression"],
    ) -> "FNode":
        new_qsimplifier = QuantifierSimplifier(self._env, self._problem)
        assert self._variable_assignments is not None
        assert self._assignments is not None
        copy = self._variable_assignments.copy()
        copy.update(variables_assignments)
        r = new_qsimplifier.qsimplify(expression, self._assignments, copy)
        assert r.is_constant()
        return r

    def walk_exists(self, expression: "FNode", args: List["FNode"]) -> "FNode":
        assert self._problem is not None
        assert len(args) == 1
        if args[0].is_bool_constant():
            if args[0].bool_constant_value():
                return self.manager.TRUE()
            return self.manager.FALSE()
        vars = expression.variables()
        type_list = [v.type for v in vars]
        possible_objects: List[List["up.model.object.Object"]] = [
            list(self._problem.objects(t)) for t in type_list
        ]
        # product of n iterables returns a generator of tuples where
        # every tuple has n elements and the tuples make every possible
        # combination of 1 item for each iterable. For example:
        # product([1,2], [3,4], [5,6], [7]) =
        # (1,3,5,7) (1,3,6,7) (1,4,5,7) (1,4,6,7) (2,3,5,7) (2,3,6,7) (2,4,5,7) (2,4,6,7)
        for o in product(*possible_objects):
            subs: Dict["Expression", "Expression"] = dict(zip(vars, list(o)))
            result = self._deep_subs_simplify(args[0], subs)
            assert result.is_bool_constant()
            if result.bool_constant_value():
                return self.manager.TRUE()
        return self.manager.FALSE()

    def walk_forall(self, expression: "FNode", args: List["FNode"]) -> "FNode":
        assert self._problem is not None
        assert len(args) == 1
        if args[0].is_bool_constant():
            if args[0].bool_constant_value():
                return self.manager.TRUE()
            return self.manager.FALSE()
        vars = expression.variables()
        type_list = [v.type for v in vars]
        possible_objects: List[List["up.model.object.Object"]] = [
            list(self._problem.objects(t)) for t in type_list
        ]
        # product of n iterables returns a generator of tuples where
        # every tuple has n elements and the tuples make every possible
        # combination of 1 item for each iterable. For example:
        # product([1,2], [3,4], [5,6], [7]) =
        # (1,3,5,7) (1,3,6,7) (1,4,5,7) (1,4,6,7) (2,3,5,7) (2,3,6,7) (2,4,5,7) (2,4,6,7)
        for o in product(*possible_objects):
            subs: Dict["Expression", "Expression"] = dict(zip(vars, list(o)))
            result = self._deep_subs_simplify(args[0], subs)
            assert result.is_bool_constant()
            if not result.bool_constant_value():
                return self.manager.FALSE()
        return self.manager.TRUE()

    def walk_fluent_exp(self, expression: "FNode", args: List["FNode"]) -> "FNode":
        new_exp = self.manager.FluentExp(expression.fluent(), tuple(args))
        assert self._assignments is not None
        res = self._assignments.get(new_exp, None)
        if res is not None:
            (res,) = self.manager.auto_promote(res)
            assert type(res) is FNode
            return res
        else:
            raise UPProblemDefinitionError(
                f"Value of Fluent {str(expression)} not found in {str(self._assignments)}"
            )

    def walk_variable_exp(self, expression: "FNode", args: List["FNode"]) -> "FNode":
        assert self._variable_assignments is not None
        res = self._variable_assignments.get(expression.variable(), None)
        if res is not None:
            (res,) = self.manager.auto_promote(res)
            assert type(res) is FNode
            return res
        else:
            raise UPProblemDefinitionError(
                f"Value of Variable {str(expression)} not found in {str(self._variable_assignments)}"
            )

    def walk_param_exp(self, expression: "FNode", args: List["FNode"]) -> "FNode":
        assert self._assignments is not None
        res = self._assignments.get(expression.parameter(), None)
        if res is not None:
            (res,) = self.manager.auto_promote(res)
            assert type(res) is FNode
            return res
        else:
            raise UPProblemDefinitionError(
                f"Value of Parameter {str(expression)} not found in {str(self._assignments)}"
            )



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from shutil import move
import warnings
import unified_planning as up
from unified_planning.shortcuts import *
from unified_planning.model.problem_kind import (
    basic_classical_kind,
    classical_kind,
    basic_numeric_kind,
    quality_metrics_kind,
    oversubscription_kind,
)
from unified_planning.test import TestCase, main, skipIfEngineNotAvailable
from unified_planning.test import skipIfNoOneshotPlannerForProblemKind
from unified_planning.test import skipIfNoOneshotPlannerSatisfiesOptimalityGuarantee
from unified_planning.test.examples import get_example_problems
from unified_planning.engines import PlanGenerationResultStatus, CompilationKind
from unified_planning.engines.results import POSITIVE_OUTCOMES
from unified_planning.model.metrics import MinimizeSequentialPlanLength


class TestPlanner(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()

    @skipIfEngineNotAvailable("tamer")
    def test_basic(self):
        problem = self.problems["basic"].problem
        a = problem.action("a")

        with OneshotPlanner(name="tamer", params={"weight": 0.8}) as planner:
            self.assertNotEqual(planner, None)
            final_report = planner.solve(problem)
            plan = final_report.plan
            self.assertEqual(
                final_report.status, PlanGenerationResultStatus.SOLVED_SATISFICING
            )
            self.assertEqual(len(plan.actions), 1)
            self.assertEqual(plan.actions[0].action, a)
            self.assertEqual(len(plan.actions[0].actual_parameters), 0)

    @skipIfEngineNotAvailable("tamer")
    def test_basic_with_timeout(self):
        problem = self.problems["basic"].problem
        a = problem.action("a")

        with OneshotPlanner(name="tamer", params={"weight": 0.8}) as planner:
            self.assertNotEqual(planner, None)
            with warnings.catch_warnings(record=True) as w:
                final_report = planner.solve(problem, timeout=0.001)
                self.assertIn(final_report.status, POSITIVE_OUTCOMES)
                plan = final_report.plan
                self.assertEqual(
                    final_report.status, PlanGenerationResultStatus.SOLVED_SATISFICING
                )
                self.assertEqual(len(plan.actions), 1)
                self.assertEqual(plan.actions[0].action, a)
                self.assertEqual(len(plan.actions[0].actual_parameters), 0)
                self.assertEqual(len(w), 1)
                self.assertEqual("Tamer does not support timeout.", str(w[-1].message))

    @skipIfEngineNotAvailable("tamer")
    def test_basic_parallel(self):
        problem = self.problems["basic"].problem
        a = problem.action("a")

        with OneshotPlanner(
            names=["tamer", "tamer"],
            params=[{"heuristic": "hadd"}, {"heuristic": "hmax"}],
        ) as planner:
            self.assertNotEqual(planner, None)
            final_report = planner.solve(problem)
            plan = final_report.plan
            self.assertEqual(
                final_report.status, PlanGenerationResultStatus.SOLVED_SATISFICING
            )
            self.assertEqual(len(plan.actions), 1)
            self.assertEqual(plan.actions[0].action, a)
            self.assertEqual(len(plan.actions[0].actual_parameters), 0)

    @skipIfNoOneshotPlannerForProblemKind(
        basic_classical_kind.union(oversubscription_kind)
    )
    def test_basic_oversubscription(self):
        problem = self.problems["basic_oversubscription"].problem
        a = problem.action("a")

        with OneshotPlanner(problem_kind=problem.kind) as planner:
            self.assertNotEqual(planner, None)
            final_report = planner.solve(problem)
            plan = final_report.plan
            self.assertEqual(
                final_report.status, PlanGenerationResultStatus.SOLVED_OPTIMALLY
            )
            self.assertEqual(len(plan.actions), 1)
            self.assertEqual(plan.actions[0].action, a)
            self.assertEqual(len(plan.actions[0].actual_parameters), 0)

    @skipIfEngineNotAvailable("tamer")
    def test_basic_oversubscription_parallel(self):
        problem = self.problems["basic_oversubscription"].problem
        a = problem.action("a")

        with OneshotPlanner(
            names=["oversubscription[tamer]", "oversubscription[tamer]"],
            params=[{"heuristic": "hadd"}, {"heuristic": "hmax"}],
        ) as planner:
            self.assertNotEqual(planner, None)
            final_report = planner.solve(problem)
            plan = final_report.plan
            self.assertEqual(
                final_report.status, PlanGenerationResultStatus.SOLVED_OPTIMALLY
            )
            self.assertEqual(len(plan.actions), 1)
            self.assertEqual(plan.actions[0].action, a)
            self.assertEqual(len(plan.actions[0].actual_parameters), 0)

    @skipIfEngineNotAvailable("tamer")
    def test_timed_connected_locations_parallel(self):
        problem = self.problems["timed_connected_locations"].problem
        move = problem.action("move")
        with OneshotPlanner(
            names=["tamer", "tamer"],
            params=[{"heuristic": "hadd"}, {"heuristic": "hmax"}],
        ) as planner:
            self.assertNotEqual(planner, None)
            with self.assertRaises(up.exceptions.UPUsageError) as e:
                final_report = planner.solve(problem)
            self.assertIn("cannot solve this kind of problem", str(e.exception))
            with Compiler(name="up_quantifiers_remover") as quantifiers_remover:
                res = quantifiers_remover.compile(
                    problem, CompilationKind.QUANTIFIERS_REMOVING
                )
                suitable_problem = res.problem
                final_report = planner.solve(suitable_problem)
                plan = final_report.plan.replace_action_instances(
                    res.map_back_action_instance
                )
                self.assertEqual(
                    final_report.status, PlanGenerationResultStatus.SOLVED_SATISFICING
                )
                self.assertEqual(len(plan.timed_actions), 2)
                self.assertEqual((plan.timed_actions[0])[1].action, move)
                self.assertEqual((plan.timed_actions[1])[1].action, move)

    @skipIfNoOneshotPlannerForProblemKind(classical_kind.union(quality_metrics_kind))
    @skipIfNoOneshotPlannerSatisfiesOptimalityGuarantee(
        PlanGenerationResultStatus.SOLVED_OPTIMALLY
    )
    def test_actions_cost(self):
        problem = self.problems["basic_with_costs"].problem
        opt_plan = self.problems["basic_with_costs"].plan
        with OneshotPlanner(
            problem_kind=problem.kind,
            optimality_guarantee=PlanGenerationResultStatus.SOLVED_OPTIMALLY,
        ) as planner:
            self.assertNotEqual(planner, None)
            final_report = planner.solve(problem)
            plan = final_report.plan
            self.assertEqual(
                final_report.status, PlanGenerationResultStatus.SOLVED_OPTIMALLY
            )
            self.assertEqual(plan, opt_plan)

    @skipIfNoOneshotPlannerForProblemKind(classical_kind.union(basic_numeric_kind))
    def test_robot(self):
        problem = self.problems["robot"].problem
        move = problem.action("move")

        with OneshotPlanner(problem_kind=problem.kind) as planner:
            self.assertNotEqual(planner, None)
            final_report = planner.solve(problem)
            plan = final_report.plan
            self.assertIn(final_report.status, POSITIVE_OUTCOMES)
            self.assertNotEqual(plan, None)
            self.assertEqual(len(plan.actions), 1)
            self.assertEqual(plan.actions[0].action, move)
            self.assertEqual(len(plan.actions[0].actual_parameters), 2)

    @skipIfNoOneshotPlannerForProblemKind(classical_kind)
    def test_robot_loader(self):
        problem = self.problems["robot_loader"].problem
        move = problem.action("move")
        load = problem.action("load")
        unload = problem.action("unload")

        with OneshotPlanner(problem_kind=problem.kind) as planner:
            self.assertNotEqual(planner, None)
            final_report = planner.solve(problem)
            plan = final_report.plan
            self.assertIn(final_report.status, POSITIVE_OUTCOMES)
            self.assertEqual(len(plan.actions), 4)
            self.assertEqual(plan.actions[0].action, move)
            self.assertEqual(plan.actions[1].action, load)
            self.assertEqual(plan.actions[2].action, move)
            self.assertEqual(plan.actions[3].action, unload)
            self.assertEqual(len(plan.actions[0].actual_parameters), 2)
            self.assertEqual(len(plan.actions[1].actual_parameters), 1)
            self.assertEqual(len(plan.actions[2].actual_parameters), 2)
            self.assertEqual(len(plan.actions[3].actual_parameters), 1)

    @skipIfNoOneshotPlannerForProblemKind(classical_kind.union(quality_metrics_kind))
    @skipIfNoOneshotPlannerSatisfiesOptimalityGuarantee(
        PlanGenerationResultStatus.SOLVED_OPTIMALLY
    )
    def test_robot_loader_adv(self):
        problem = self.problems["robot_loader_adv"].problem.clone()
        problem.add_quality_metric(MinimizeSequentialPlanLength())

        move = problem.action("move")
        load = problem.action("load")
        unload = problem.action("unload")

        with OneshotPlanner(
            problem_kind=problem.kind, optimality_guarantee="SOLVED_OPTIMALLY"
        ) as planner:
            self.assertNotEqual(planner, None)
            final_report = planner.solve(problem)
            plan = final_report.plan
            self.assertIn(final_report.status, POSITIVE_OUTCOMES)
            self.assertEqual(len(plan.actions), 5)
            self.assertEqual(plan.actions[0].action, move)
            self.assertEqual(plan.actions[1].action, load)
            self.assertEqual(plan.actions[2].action, move)
            self.assertEqual(plan.actions[3].action, unload)
            self.assertEqual(plan.actions[4].action, move)
            self.assertEqual(len(plan.actions[0].actual_parameters), 3)
            self.assertEqual(len(plan.actions[1].actual_parameters), 3)
            self.assertEqual(len(plan.actions[2].actual_parameters), 3)
            self.assertEqual(len(plan.actions[3].actual_parameters), 3)
            self.assertEqual(len(plan.actions[4].actual_parameters), 3)


if __name__ == "__main__":
    main()



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from unified_planning.model.walkers.dag import DagWalker
from unified_planning.model.walkers.generic import handles
from unified_planning.model.walkers.dnf import Dnf, Nnf
from unified_planning.model.walkers.expression_quantifiers_remover import (
    ExpressionQuantifiersRemover,
)
from unified_planning.model.walkers.operators_extractor import OperatorsExtractor
from unified_planning.model.walkers.quantifier_simplifier import QuantifierSimplifier
from unified_planning.model.walkers.simplifier import Simplifier
from unified_planning.model.walkers.state_evaluator import StateEvaluator
from unified_planning.model.walkers.substituter import Substituter
from unified_planning.model.walkers.type_checker import TypeChecker
from unified_planning.model.walkers.free_vars import FreeVarsExtractor



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unified_planning.shortcuts import *
from unified_planning.model.htn import *
from collections import namedtuple

Example = namedtuple("Example", ["problem", "plan"])


def get_example_problems():
    problems = {}

    htn = HierarchicalProblem()

    Location = UserType("Location")
    l1 = htn.add_object("l1", Location)
    l2 = htn.add_object("l2", Location)
    l3 = htn.add_object("l3", Location)
    l4 = htn.add_object("l4", Location)

    loc = htn.add_fluent("loc", Location)

    connected = Fluent("connected", l1=Location, l2=Location)
    htn.add_fluent(connected, default_initial_value=False)
    htn.set_initial_value(connected(l1, l2), True)
    htn.set_initial_value(connected(l2, l3), True)
    htn.set_initial_value(connected(l3, l4), True)
    htn.set_initial_value(connected(l4, l3), True)
    htn.set_initial_value(connected(l3, l2), True)
    htn.set_initial_value(connected(l2, l1), True)

    move = InstantaneousAction("move", l_from=Location, l_to=Location)
    l_from = move.parameter("l_from")
    l_to = move.parameter("l_to")
    move.add_precondition(Equals(loc, l_from))
    move.add_precondition(connected(l_from, l_to))
    move.add_effect(loc, l_to)
    htn.add_action(move)
    go = htn.add_task("go", target=Location)

    go_noop = Method("go-noop", target=Location)
    go_noop.set_task(go)
    target = go_noop.parameter("target")
    go_noop.add_precondition(Equals(loc, target))
    htn.add_method(go_noop)

    go_recursive = Method(
        "go-recursive", source=Location, inter=Location, target=Location
    )
    go_recursive.set_task(go, go_recursive.parameter("target"))
    source = go_recursive.parameter("source")
    inter = go_recursive.parameter("inter")
    target = go_recursive.parameter("target")
    go_recursive.add_precondition(Equals(loc, source))
    go_recursive.add_precondition(connected(source, inter))
    t1 = go_recursive.add_subtask(move, source, inter)
    t2 = go_recursive.add_subtask(go, target)
    go_recursive.set_ordered(t1, t2)
    htn.add_method(go_recursive)

    go1 = htn.task_network.add_subtask(go, l4)
    final_loc = htn.task_network.add_variable("final_loc", Location)
    go2 = htn.task_network.add_subtask(go, final_loc)
    htn.task_network.add_constraint(Or(Equals(final_loc, l1), Equals(final_loc, l2)))
    htn.task_network.set_strictly_before(go1, go2)

    htn.set_initial_value(loc, l1)
    plan = up.plans.SequentialPlan(
        [
            up.plans.ActionInstance(move, (ObjectExp(l1), ObjectExp(l2))),
            up.plans.ActionInstance(move, (ObjectExp(l2), ObjectExp(l3))),
            up.plans.ActionInstance(move, (ObjectExp(l3), ObjectExp(l4))),
            up.plans.ActionInstance(move, (ObjectExp(l4), ObjectExp(l3))),
            up.plans.ActionInstance(move, (ObjectExp(l3), ObjectExp(l2))),
        ]
    )
    htn_go = Example(problem=htn, plan=plan)

    problems["htn-go"] = htn_go

    return problems


if __name__ == "__main__":
    for name, problem in get_example_problems().items():
        print(f"======= {name} ======")
        print(str(problem))



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This module defines the PlanGenerationResult class."""


import unified_planning as up
from unified_planning.exceptions import UPUsageError
from unified_planning.model import Problem
from unified_planning.plans import ActionInstance
from dataclasses import dataclass, field
from enum import Enum, auto
from typing import Callable, Dict, Optional, List


class ValidationResultStatus(Enum):
    VALID = (
        auto()
    )  # The plan is valid for the problem, it satisfies all the hard constraints
    INVALID = (
        auto()
    )  # The plan is invalid for the problem, it does not satisfy all the hard constraints


class PlanGenerationResultStatus(Enum):
    SOLVED_SATISFICING = auto()  # Valid plan found.
    SOLVED_OPTIMALLY = auto()  # Optimal plan found.
    UNSOLVABLE_PROVEN = auto()  # The problem is impossible, no valid plan exists.
    UNSOLVABLE_INCOMPLETELY = (
        auto()
    )  # The planner could not find a plan, but it's not sure that the problem is impossible (The planner is incomplete)
    TIMEOUT = auto()  # The planner ran out of time
    MEMOUT = auto()  # The planner ran out of memory
    INTERNAL_ERROR = auto()  # The planner had an internal error
    UNSUPPORTED_PROBLEM = auto()  # The problem given is not supported by the planner
    INTERMEDIATE = (
        auto()
    )  # The report is not a final one but it's given through the callback function


POSITIVE_OUTCOMES = frozenset(
    [
        PlanGenerationResultStatus.SOLVED_SATISFICING,
        PlanGenerationResultStatus.SOLVED_OPTIMALLY,
    ]
)

NEGATIVE_OUTCOMES = frozenset(
    [
        PlanGenerationResultStatus.UNSOLVABLE_PROVEN,
        PlanGenerationResultStatus.UNSOLVABLE_INCOMPLETELY,
        PlanGenerationResultStatus.UNSUPPORTED_PROBLEM,
    ]
)


class LogLevel(Enum):
    DEBUG = auto()
    INFO = auto()
    WARNING = auto()
    ERROR = auto()


@dataclass
class LogMessage:
    """This class is composed by a message and an integer indicating this message level, like Debug, Info, Warning or Error."""

    level: LogLevel
    message: str


@dataclass
class Result:
    """This class represents the base class for results given by the engines to the user."""

    def is_definitive_result(self, *args) -> bool:
        """This predicate should state if the Result is definitive or if it can be improved."""
        raise NotImplementedError


@dataclass
class PlanGenerationResult(Result):
    """Class that represents the result of a plan generation call."""

    status: PlanGenerationResultStatus
    plan: Optional["up.plans.Plan"]
    engine_name: str
    metrics: Optional[Dict[str, str]] = field(default=None)
    log_messages: Optional[List[LogMessage]] = field(default=None)

    def __post__init(self):
        # Checks that plan and status are consistent
        if self.status in POSITIVE_OUTCOMES and self.plan is None:
            raise UPUsageError(
                f"The Result status is {str(self.status)} but no plan is set."
            )
        elif self.status in NEGATIVE_OUTCOMES and self.plan is not None:
            raise UPUsageError(
                f"The Result status is {str(self.status)} but the plan is {str(self.plan)}.\nWith this status the plan must be None."
            )
        return self

    def is_definitive_result(self, *args) -> bool:
        optimality_required = False
        if len(args) > 0:
            optimality_required = (
                len(args[0].quality_metrics) > 0
            )  # Require optimality if the problem has at least one quality metric.
        return (
            self.status == PlanGenerationResultStatus.SOLVED_OPTIMALLY
            or self.status == PlanGenerationResultStatus.UNSOLVABLE_PROVEN
            or (
                optimality_required
                and self.status == PlanGenerationResultStatus.SOLVED_SATISFICING
            )
        )


@dataclass
class ValidationResult(Result):
    """Class that represents the result of a validate call."""

    status: ValidationResultStatus
    engine_name: str
    log_messages: Optional[List[LogMessage]] = field(default=None)

    def is_definitive_result(self, *args) -> bool:
        return True


@dataclass
class CompilerResult(Result):
    """Class that represents the result of a compile call."""

    problem: Optional[Problem]
    map_back_action_instance: Optional[Callable[[ActionInstance], ActionInstance]]
    engine_name: str
    log_messages: Optional[List[LogMessage]] = field(default=None)

    def _post_init(self):
        # Check that compiled problem and map_back_action_instance are consistent with each other
        if self.problem is None and self.map_back_action_instance is not None:
            raise UPUsageError(
                f"The compiled Problem is None but the map_back_action_instance Callable is not None."
            )
        if self.problem is not None and self.map_back_action_instance is None:
            raise UPUsageError(
                f"The compiled Problem is {str(self.problem)} but the map_back_action_instance Callable is None."
            )

    def is_definitive_result(self, *args) -> bool:
        return self.problem is not None



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unified_planning as up
import unified_planning.engines
from unified_planning.model.problem_kind import ProblemKind
from unified_planning.environment import get_env
from typing import List, Optional, Union


FILE_PATH = os.path.dirname(os.path.abspath(__file__))


class ENHSP(up.engines.PDDLPlanner):
    def __init__(self):
        up.engines.PDDLPlanner.__init__(self, False)

    @property
    def name(self) -> str:
        return "ENHSP"

    def _get_cmd(
        self, domanin_filename: str, problem_filename: str, plan_filename: str
    ) -> List[str]:
        return [
            "java",
            "-jar",
            os.path.join(
                FILE_PATH, "..", "..", "..", ".planners", "enhsp-20", "enhsp.jar"
            ),
            "-o",
            domanin_filename,
            "-f",
            problem_filename,
            "-sp",
            plan_filename,
            "-planner",
            "opt-hrmax",
        ]

    def _result_status(
        self, problem: "up.model.Problem", plan: Optional["up.plans.Plan"]
    ) -> "up.engines.results.PlanGenerationResultStatus":
        if plan is None:
            return up.engines.results.PlanGenerationResultStatus.UNSOLVABLE_PROVEN
        else:
            return up.engines.results.PlanGenerationResultStatus.SOLVED_OPTIMALLY

    @staticmethod
    def satisfies(optimality_guarantee: up.engines.OptimalityGuarantee) -> bool:
        return True

    @staticmethod
    def supported_kind() -> ProblemKind:
        supported_kind = ProblemKind()
        supported_kind.set_problem_class("ACTION_BASED")
        supported_kind.set_numbers("DISCRETE_NUMBERS")
        supported_kind.set_numbers("CONTINUOUS_NUMBERS")
        supported_kind.set_typing("FLAT_TYPING")
        supported_kind.set_typing("HIERARCHICAL_TYPING")
        supported_kind.set_conditions_kind("NEGATIVE_CONDITIONS")
        supported_kind.set_conditions_kind("DISJUNCTIVE_CONDITIONS")
        supported_kind.set_conditions_kind("EQUALITY")
        supported_kind.set_conditions_kind("EXISTENTIAL_CONDITIONS")
        supported_kind.set_conditions_kind("UNIVERSAL_CONDITIONS")
        supported_kind.set_effects_kind("CONDITIONAL_EFFECTS")
        supported_kind.set_effects_kind("INCREASE_EFFECTS")
        supported_kind.set_effects_kind("DECREASE_EFFECTS")
        supported_kind.set_fluents_type("NUMERIC_FLUENTS")
        supported_kind.set_quality_metrics("ACTIONS_COST")
        supported_kind.set_quality_metrics("FINAL_VALUE")
        supported_kind.set_quality_metrics("PLAN_LENGTH")
        return supported_kind

    @staticmethod
    def supports(problem_kind: "ProblemKind") -> bool:
        return problem_kind <= ENHSP.supported_kind()

    @staticmethod
    def get_credits(**kwargs) -> Optional[up.engines.Credits]:
        return None


env = get_env()
if os.path.isfile(
    os.path.join(FILE_PATH, "..", "..", "..", ".planners", "enhsp-20", "enhsp.jar")
):
    env.factory.add_engine(
        "opt-pddl-planner", "unified_planning.test.pddl.enhsp", "ENHSP"
    )



from unified_planning.io.python_writer import PythonWriter
from unified_planning.io.pddl_reader import PDDLReader
from unified_planning.io.pddl_writer import PDDLWriter
from unified_planning.io.anml_writer import ANMLWriter



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This module defines all the operators used by the Unified Planning library."""
from enum import Enum, auto


class OperatorKind(Enum):
    AND = auto()
    OR = auto()
    NOT = auto()
    IMPLIES = auto()
    IFF = auto()
    EXISTS = auto()
    FORALL = auto()
    FLUENT_EXP = auto()
    PARAM_EXP = auto()
    VARIABLE_EXP = auto()
    OBJECT_EXP = auto()
    TIMING_EXP = auto()
    BOOL_CONSTANT = auto()
    INT_CONSTANT = auto()
    REAL_CONSTANT = auto()
    PLUS = auto()
    MINUS = auto()
    TIMES = auto()
    DIV = auto()
    LE = auto()
    LT = auto()
    EQUALS = auto()


BOOL_OPERATORS = frozenset(
    [
        OperatorKind.AND,
        OperatorKind.OR,
        OperatorKind.NOT,
        OperatorKind.IMPLIES,
        OperatorKind.IFF,
        OperatorKind.EXISTS,
        OperatorKind.FORALL,
    ]
)

CONSTANTS = frozenset(
    [OperatorKind.BOOL_CONSTANT, OperatorKind.REAL_CONSTANT, OperatorKind.INT_CONSTANT]
)

IRA_RELATIONS = frozenset([OperatorKind.LE, OperatorKind.LT])

RELATIONS = frozenset((OperatorKind.EQUALS,)) | IRA_RELATIONS

IRA_OPERATORS = frozenset(
    [OperatorKind.PLUS, OperatorKind.MINUS, OperatorKind.TIMES, OperatorKind.DIV]
)



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import warnings
import unified_planning as up
import unified_planning.engines as engines
from unified_planning.plans import Plan
from unified_planning.model import ProblemKind
from unified_planning.exceptions import UPUsageError
from unified_planning.engines.results import (
    LogLevel,
    PlanGenerationResultStatus,
    Result,
    ValidationResult,
    PlanGenerationResult,
)
from typing import IO, Callable, Dict, List, Optional, Tuple, Type, cast
from fractions import Fraction
from multiprocessing import Process, Queue


class Parallel(
    engines.engine.Engine,
    engines.mixins.OneshotPlannerMixin,
    engines.mixins.PlanValidatorMixin,
):
    """Create a parallel instance of multiple Engines."""

    def __init__(
        self,
        factory: "up.engines.factory.Factory",
        engines: List[Tuple[str, Dict[str, str]]],
    ):
        self.engines = engines
        self._factory = factory

    @property
    def name(self) -> str:
        return "Parallel"

    @staticmethod
    def supports(problem_kind: "ProblemKind") -> bool:
        # The supported features depends on its actual engines
        return True

    def _run_parallel(self, fname, *args) -> List[Result]:
        signaling_queue: Queue = Queue()
        processes = []
        for idx, (engine_name, opts) in enumerate(self.engines):
            options = opts
            _p = Process(
                name=str(idx),
                target=_run,
                args=(
                    idx,
                    self._factory,
                    engine_name,
                    options,
                    signaling_queue,
                    fname,
                    *args,
                ),
            )
            processes.append(_p)
            _p.start()
        processes_alive = len(processes)
        results: List[Result] = []
        definitive_result_found: bool = False
        while True:
            if processes_alive == 0:  # Every planner gave a result
                break
            (idx, res) = signaling_queue.get(block=True)
            processes_alive -= 1
            if isinstance(res, BaseException):
                raise res
            else:
                assert isinstance(res, Result)
                # If the planner is sure about the result (optimality of the result or impossibility of the problem or the problem does not need optimality) exit the loop
                if res.is_definitive_result(*args):
                    definitive_result_found = True
                    break
                else:
                    results.append(res)
        for p in processes:
            p.terminate()
        if definitive_result_found:  # A planner found a definitive result
            return [res]
        return results

    def _solve(
        self,
        problem: "up.model.AbstractProblem",
        callback: Optional[
            Callable[["up.engines.results.PlanGenerationResult"], None]
        ] = None,
        timeout: Optional[float] = None,
        output_stream: Optional[IO[str]] = None,
    ) -> "up.engines.results.PlanGenerationResult":
        for engine_name, _ in self.engines:
            engine = self._factory.engine(engine_name)
            assert issubclass(engine, engines.mixins.OneshotPlannerMixin)
            if not engine.supports(problem.kind):
                raise UPUsageError(
                    "Parallel engines cannot solve this kind of problem!"
                )
        if callback is not None:
            warnings.warn(
                "Parallel engines do not support the callback system.", UserWarning
            )
        if output_stream is not None:
            warnings.warn(
                "Parallel engines do not support the output stream system.", UserWarning
            )

        final_reports = self._run_parallel("solve", problem, None, timeout, None)

        result_order: List[PlanGenerationResultStatus] = [
            PlanGenerationResultStatus.SOLVED_OPTIMALLY,  # List containing the results in the order we prefer them
            PlanGenerationResultStatus.UNSOLVABLE_PROVEN,
            PlanGenerationResultStatus.SOLVED_SATISFICING,
            PlanGenerationResultStatus.UNSOLVABLE_INCOMPLETELY,
            PlanGenerationResultStatus.TIMEOUT,
            PlanGenerationResultStatus.MEMOUT,
            PlanGenerationResultStatus.INTERNAL_ERROR,
            PlanGenerationResultStatus.UNSUPPORTED_PROBLEM,
        ]
        final_result: Optional[PlanGenerationResult] = None
        result_found: bool = False
        for ro in result_order:
            if result_found:
                break
            for r in final_reports:
                pgr = cast(PlanGenerationResult, r)
                if pgr.status == ro:
                    result_found = True
                    final_result = pgr
                    break
        logs = [up.engines.LogMessage(LogLevel.INFO, str(fr)) for fr in final_reports]
        # if no results are given by the planner, we create a default one
        if final_result is None:
            return up.engines.PlanGenerationResult(
                PlanGenerationResultStatus.UNSOLVABLE_INCOMPLETELY,
                None,
                self.name,
                log_messages=logs,
            )
        new_plan = (
            problem.normalize_plan(final_result.plan)
            if final_result.plan is not None
            else None
        )
        if final_result.log_messages is not None:
            logs = final_result.log_messages + logs
        return up.engines.results.PlanGenerationResult(
            final_result.status,
            new_plan,
            final_result.engine_name,
            final_result.metrics,
            logs,
        )

    def _validate(
        self, problem: "up.model.AbstractProblem", plan: Plan
    ) -> "up.engines.results.ValidationResult":
        for engine_name, _ in self.engines:
            engine = self._factory.engine(engine_name)
            assert issubclass(engine, engines.mixins.PlanValidatorMixin)
            if not engine.supports(problem.kind):
                raise UPUsageError(
                    "Parallel engines cannot validate this kind of problem!"
                )
        return cast(ValidationResult, self._run_parallel("validate", problem, plan)[0])


def _run(
    idx: int,
    factory: "up.engines.factory.Factory",
    engine_name: str,
    options: Dict[str, str],
    signaling_queue: Queue,
    fname: str,
    *args,
):
    EngineClass = factory.engine(engine_name)
    with EngineClass(**options) as s:
        try:
            local_res = getattr(s, fname)(*args)
        except Exception as ex:
            signaling_queue.put((idx, ex))
            return
        signaling_queue.put((idx, local_res))



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from unified_planning.engines.engine import Engine
from unified_planning.engines.meta_engine import MetaEngine
from unified_planning.engines.credits import Credits
from unified_planning.engines.factory import Factory
from unified_planning.engines.parallel import Parallel
from unified_planning.engines.pddl_planner import PDDLPlanner
from unified_planning.engines.plan_validator import SequentialPlanValidator
from unified_planning.engines.oversubscription_planner import OversubscriptionPlanner
from unified_planning.engines.results import (
    Result,
    LogMessage,
    PlanGenerationResult,
    LogLevel,
    PlanGenerationResultStatus,
    ValidationResult,
    ValidationResultStatus,
    CompilerResult,
)
from unified_planning.engines.sequential_simulator import (
    SequentialSimulator,
    InstantaneousEvent,
)
from unified_planning.engines.mixins.simulator import SimulatorMixin, Event
from unified_planning.engines.mixins.oneshot_planner import OptimalityGuarantee
from unified_planning.engines.mixins.compiler import CompilationKind

__all__ = [
    "Factory",
    "Grounder",
    "Parallel",
    "PDDLPlanner",
    "SequentialPlanValidator",
    "SimulatorMixin",
    "SequentialSimulator",
    "Event",
    "InstantaneousEvent",
    "Engine",
    "OptimalityGuarantee",
    "CompilationKind",
    "Credits",
    "Result",
    "LogMessage",
    "PlanGenerationResult",
    "LogLevel",
    "PlanGenerationResultStatus",
    "ValidationResult",
    "ValidationResultStatus",
    "CompilerResult",
]



from unified_planning.model.htn.hierarchical_problem import HierarchicalProblem
from unified_planning.model.htn.method import Method, ParameterizedTask
from unified_planning.model.htn.task import Task, Subtask
from unified_planning.model.htn.task_network import TaskNetwork

__all__ = [
    "Method",
    "Task",
    "HierarchicalProblem",
    "Subtask",
    "TaskNetwork",
    "ParameterizedTask",
]



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unified_planning as up
from unified_planning.shortcuts import *
from unified_planning.test import TestCase, main, examples
from unified_planning.test.examples import get_example_problems


class TestProblem(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()

    def test_problem_kind(self):
        problem_kind = ProblemKind()
        self.assertFalse(problem_kind.has_discrete_time())
        self.assertFalse(problem_kind.has_continuous_time())
        problem_kind.set_time("DISCRETE_TIME")
        self.assertTrue(problem_kind.has_discrete_time())
        problem_kind.set_time("CONTINUOUS_TIME")
        self.assertTrue(problem_kind.has_continuous_time())

    def test_basic(self):
        problem = self.problems["basic"].problem

        x = problem.fluent("x")
        self.assertEqual(x.name, "x")
        self.assertEqual(str(x), "bool x")
        self.assertEqual(x.arity, 0)
        self.assertTrue(x.type.is_bool_type())

        a = problem.action("a")
        self.assertEqual(a.name, "a")
        self.assertEqual(len(a.preconditions), 1)
        self.assertEqual(len(a.effects), 1)
        a_str = str(a)
        self.assertIn("action a", a_str)
        self.assertIn("preconditions", a_str)
        self.assertIn("not x", a_str)
        self.assertIn("effects", a_str)
        self.assertIn("x := true", a_str)

        self.assertEqual(problem.name, "basic")
        self.assertEqual(len(problem.fluents), 1)
        self.assertEqual(len(problem.actions), 1)
        self.assertTrue(problem.initial_value(x) is not None)
        self.assertEqual(len(problem.goals), 1)
        problem_str = str(problem)
        self.assertIn("fluents", problem_str)
        self.assertIn("actions", problem_str)
        self.assertIn("initial values", problem_str)
        self.assertIn("goals", problem_str)

    def test_basic_conditional(self):
        problem = self.problems["basic_conditional"].problem

        x = problem.fluent("x")
        self.assertEqual(x.name, "x")
        self.assertEqual(str(x), "bool x")
        self.assertEqual(x.arity, 0)
        self.assertTrue(x.type.is_bool_type())

        y = problem.fluent("y")
        self.assertEqual(y.name, "y")
        self.assertEqual(str(y), "bool y")
        self.assertEqual(y.arity, 0)
        self.assertTrue(y.type.is_bool_type())

        a_x = problem.action("a_x")
        self.assertEqual(a_x.name, "a_x")
        self.assertEqual(len(a_x.preconditions), 1)
        self.assertEqual(len(a_x.effects), 1)
        ax_str = str(a_x)
        self.assertIn("action a_x", ax_str)
        self.assertIn("preconditions", ax_str)
        self.assertIn("not x", ax_str)
        self.assertIn("effects", ax_str)
        self.assertIn("if y then x := true", ax_str)

        a_y = problem.action("a_y")
        self.assertEqual(a_y.name, "a_y")
        self.assertEqual(len(a_y.preconditions), 1)
        self.assertEqual(len(a_y.effects), 1)
        ay_str = str(a_y)
        self.assertIn("action a_y", ay_str)
        self.assertIn("preconditions", ay_str)
        self.assertIn("not y", ay_str)
        self.assertIn("effects", ay_str)
        self.assertIn("y := true", ay_str)

        self.assertEqual(problem.name, "basic_conditional")
        self.assertEqual(len(problem.fluents), 2)
        self.assertEqual(len(problem.actions), 2)
        self.assertTrue(problem.initial_value(x) is not None)
        self.assertTrue(problem.initial_value(y) is not None)
        self.assertEqual(len(problem.goals), 1)
        problem_str = str(problem)
        self.assertIn("fluents", problem_str)
        self.assertIn("actions", problem_str)
        self.assertIn("initial values", problem_str)
        self.assertIn("goals", problem_str)

    def test_robot(self):
        problem = self.problems["robot"].problem

        Location = problem.user_type("Location")
        self.assertTrue(Location.is_user_type())
        self.assertEqual(Location.name, "Location")
        self.assertEqual(str(Location), "Location")

        robot_at = problem.fluent("robot_at")
        self.assertEqual(robot_at.name, "robot_at")
        self.assertEqual(str(robot_at), "bool robot_at[position=Location]")
        self.assertEqual(robot_at.arity, 1)
        self.assertEqual(
            robot_at.signature, [up.model.Parameter("position", Location, problem.env)]
        )
        self.assertTrue(robot_at.type.is_bool_type())

        battery_charge = problem.fluent("battery_charge")
        self.assertEqual(battery_charge.name, "battery_charge")
        self.assertEqual(str(battery_charge), "real[0, 100] battery_charge")
        self.assertEqual(battery_charge.arity, 0)
        self.assertTrue(battery_charge.type.is_real_type())

        move = problem.action("move")
        l_from = move.parameter("l_from")
        l_to = move.parameter("l_to")
        self.assertEqual(move.name, "move")
        self.assertEqual(len(move.parameters), 2)
        self.assertEqual(l_from.name, "l_from")
        self.assertEqual(l_from.type, Location)
        self.assertEqual(l_to.name, "l_to")
        self.assertEqual(l_to.type, Location)
        self.assertEqual(len(move.preconditions), 4)
        self.assertEqual(len(move.effects), 3)
        move_str = str(move)
        self.assertTrue("action move(Location l_from, Location l_to)" in move_str)
        self.assertTrue("preconditions" in move_str)
        self.assertTrue("10 <= battery_charge" in move_str)
        self.assertTrue("not (l_from == l_to)" in move_str)
        self.assertTrue("robot_at(l_from)" in move_str)
        self.assertTrue("not robot_at(l_to)" in move_str)
        self.assertTrue("effects" in move_str)
        self.assertTrue("robot_at(l_from) := false" in move_str)
        self.assertTrue("robot_at(l_to) := true" in move_str)
        self.assertTrue("battery_charge := (battery_charge - 10)" in move_str)

        l1 = problem.object("l1")
        l2 = problem.object("l2")
        self.assertEqual(l1.name, "l1")
        self.assertEqual(str(l1), "l1")
        self.assertEqual(l1.type, Location)
        self.assertEqual(l2.name, "l2")
        self.assertEqual(str(l2), "l2")
        self.assertEqual(l2.type, Location)

        self.assertEqual(problem.name, "robot")
        self.assertEqual(len(problem.fluents), 2)
        self.assertEqual(problem.fluent("robot_at"), robot_at)
        self.assertEqual(problem.fluent("battery_charge"), battery_charge)
        self.assertEqual(len(problem.user_types), 1)
        self.assertEqual(problem.user_type("Location"), Location)
        self.assertEqual(len(list(problem.objects(Location))), 2)
        self.assertEqual(list(problem.objects(Location)), [l1, l2])
        self.assertEqual(len(problem.actions), 1)
        self.assertEqual(problem.action("move"), move)
        self.assertTrue(problem.initial_value(robot_at(l1)) is not None)
        self.assertTrue(problem.initial_value(robot_at(l2)) is not None)
        self.assertTrue(problem.initial_value(battery_charge) is not None)
        self.assertEqual(len(problem.goals), 1)
        problem_str = str(problem)
        self.assertTrue("types" in problem_str)
        self.assertTrue("fluents" in problem_str)
        self.assertTrue("actions" in problem_str)
        self.assertTrue("objects" in problem_str)
        self.assertTrue("initial values" in problem_str)
        self.assertTrue("goals" in problem_str)

    def test_robot_loader(self):
        problem = self.problems["robot_loader"].problem

        Location = problem.user_type("Location")
        self.assertTrue(Location.is_user_type())
        self.assertEqual(Location.name, "Location")

        robot_at = problem.fluent("robot_at")
        self.assertEqual(robot_at.name, "robot_at")
        self.assertEqual(robot_at.arity, 1)
        self.assertEqual(
            robot_at.signature, [up.model.Parameter("position", Location, problem.env)]
        )
        self.assertTrue(robot_at.type.is_bool_type())

        cargo_at = problem.fluent("cargo_at")
        self.assertEqual(cargo_at.name, "cargo_at")
        self.assertEqual(cargo_at.arity, 1)
        self.assertEqual(
            cargo_at.signature, [up.model.Parameter("position", Location, problem.env)]
        )
        self.assertTrue(cargo_at.type.is_bool_type())

        cargo_mounted = problem.fluent("cargo_mounted")
        self.assertEqual(cargo_mounted.name, "cargo_mounted")
        self.assertEqual(cargo_mounted.arity, 0)
        self.assertTrue(cargo_mounted.type.is_bool_type())

        move = problem.action("move")
        l_from = move.parameter("l_from")
        l_to = move.parameter("l_to")
        self.assertEqual(move.name, "move")
        self.assertEqual(len(move.parameters), 2)
        self.assertEqual(l_from.name, "l_from")
        self.assertEqual(l_from.type, Location)
        self.assertEqual(l_to.name, "l_to")
        self.assertEqual(l_to.type, Location)
        self.assertEqual(len(move.preconditions), 3)
        self.assertEqual(len(move.effects), 2)

        load = problem.action("load")
        loc = load.parameter("loc")
        self.assertEqual(load.name, "load")
        self.assertEqual(len(load.parameters), 1)
        self.assertEqual(loc.name, "loc")
        self.assertEqual(loc.type, Location)
        self.assertEqual(len(load.preconditions), 3)
        self.assertEqual(len(load.effects), 2)

        unload = problem.action("unload")
        loc = unload.parameter("loc")
        self.assertEqual(unload.name, "unload")
        self.assertEqual(len(unload.parameters), 1)
        self.assertEqual(loc.name, "loc")
        self.assertEqual(loc.type, Location)
        self.assertEqual(len(unload.preconditions), 3)
        self.assertEqual(len(unload.effects), 2)

        l1 = problem.object("l1")
        l2 = problem.object("l2")
        self.assertEqual(l1.name, "l1")
        self.assertEqual(l1.type, Location)
        self.assertEqual(l2.name, "l2")
        self.assertEqual(l2.type, Location)

        self.assertEqual(problem.name, "robot_loader")
        self.assertEqual(len(problem.fluents), 3)
        self.assertEqual(problem.fluent("robot_at"), robot_at)
        self.assertEqual(problem.fluent("cargo_at"), cargo_at)
        self.assertEqual(problem.fluent("cargo_mounted"), cargo_mounted)
        self.assertEqual(len(problem.user_types), 1)
        self.assertEqual(problem.user_type("Location"), Location)
        self.assertEqual(len(list(problem.objects(Location))), 2)
        self.assertEqual(list(problem.objects(Location)), [l1, l2])
        self.assertEqual(len(problem.actions), 3)
        self.assertEqual(problem.action("move"), move)
        self.assertEqual(problem.action("load"), load)
        self.assertEqual(problem.action("unload"), unload)
        self.assertTrue(problem.initial_value(robot_at(l1)) is not None)
        self.assertTrue(problem.initial_value(robot_at(l2)) is not None)
        self.assertTrue(problem.initial_value(cargo_at(l1)) is not None)
        self.assertTrue(problem.initial_value(cargo_at(l2)) is not None)
        self.assertTrue(problem.initial_value(cargo_mounted) is not None)
        self.assertEqual(len(problem.goals), 1)

    def test_robot_loader_adv(self):
        problem = self.problems["robot_loader_adv"].problem

        Location = problem.user_type("Location")
        self.assertTrue(Location.is_user_type())
        self.assertEqual(Location.name, "Location")

        Robot = problem.user_type("Robot")
        self.assertTrue(Robot.is_user_type())
        self.assertEqual(Robot.name, "Robot")

        Container = problem.user_type("Container")
        self.assertTrue(Container.is_user_type())
        self.assertEqual(Container.name, "Container")

        robot_at = problem.fluent("robot_at")
        self.assertEqual(robot_at.name, "robot_at")
        self.assertEqual(robot_at.arity, 2)
        self.assertEqual(
            robot_at.signature,
            [
                up.model.Parameter("robot", Robot, problem.env),
                up.model.Parameter("position", Location, problem.env),
            ],
        )
        self.assertTrue(robot_at.type.is_bool_type())

        cargo_at = problem.fluent("cargo_at")
        self.assertEqual(cargo_at.name, "cargo_at")
        self.assertEqual(cargo_at.arity, 2)
        self.assertEqual(
            cargo_at.signature,
            [
                up.model.Parameter("cargo", Container, problem.env),
                up.model.Parameter("position", Location, problem.env),
            ],
        )
        self.assertTrue(cargo_at.type.is_bool_type())

        cargo_mounted = problem.fluent("cargo_mounted")
        self.assertEqual(cargo_mounted.name, "cargo_mounted")
        self.assertEqual(cargo_mounted.arity, 2)
        self.assertEqual(
            cargo_mounted.signature,
            [
                up.model.Parameter("cargo", Container, problem.env),
                up.model.Parameter("robot", Robot, problem.env),
            ],
        )
        self.assertTrue(cargo_mounted.type.is_bool_type())

        move = problem.action("move")
        l_from = move.parameter("l_from")
        l_to = move.parameter("l_to")
        r = move.parameter("r")
        self.assertEqual(move.name, "move")
        self.assertEqual(len(move.parameters), 3)
        self.assertEqual(l_from.name, "l_from")
        self.assertEqual(l_from.type, Location)
        self.assertEqual(l_to.name, "l_to")
        self.assertEqual(l_to.type, Location)
        self.assertEqual(r.name, "r")
        self.assertEqual(r.type, Robot)
        self.assertEqual(len(move.preconditions), 3)
        self.assertEqual(len(move.effects), 2)

        load = problem.action("load")
        loc = load.parameter("loc")
        r = load.parameter("r")
        c = load.parameter("c")
        self.assertEqual(load.name, "load")
        self.assertEqual(len(load.parameters), 3)
        self.assertEqual(loc.name, "loc")
        self.assertEqual(loc.type, Location)
        self.assertEqual(r.name, "r")
        self.assertEqual(r.type, Robot)
        self.assertEqual(c.name, "c")
        self.assertEqual(c.type, Container)
        self.assertEqual(len(load.preconditions), 3)
        self.assertEqual(len(load.effects), 2)

        unload = problem.action("unload")
        loc = unload.parameter("loc")
        r = unload.parameter("r")
        c = unload.parameter("c")
        self.assertEqual(unload.name, "unload")
        self.assertEqual(len(unload.parameters), 3)
        self.assertEqual(loc.name, "loc")
        self.assertEqual(loc.type, Location)
        self.assertEqual(r.name, "r")
        self.assertEqual(r.type, Robot)
        self.assertEqual(c.name, "c")
        self.assertEqual(c.type, Container)
        self.assertEqual(len(unload.preconditions), 3)
        self.assertEqual(len(unload.effects), 2)

        l1 = problem.object("l1")
        l2 = problem.object("l2")
        l3 = problem.object("l3")
        r1 = problem.object("r1")
        c1 = problem.object("c1")
        self.assertEqual(l1.name, "l1")
        self.assertEqual(l1.type, Location)
        self.assertEqual(l2.name, "l2")
        self.assertEqual(l2.type, Location)
        self.assertEqual(l3.name, "l3")
        self.assertEqual(l3.type, Location)
        self.assertEqual(r1.name, "r1")
        self.assertEqual(r1.type, Robot)
        self.assertEqual(c1.name, "c1")
        self.assertEqual(c1.type, Container)

        self.assertEqual(problem.name, "robot_loader_adv")
        self.assertEqual(len(problem.fluents), 3)
        self.assertEqual(problem.fluent("robot_at"), robot_at)
        self.assertEqual(problem.fluent("cargo_at"), cargo_at)
        self.assertEqual(problem.fluent("cargo_mounted"), cargo_mounted)
        self.assertEqual(len(problem.user_types), 3)
        self.assertEqual(problem.user_type("Location"), Location)
        self.assertEqual(len(list(problem.objects(Location))), 3)
        self.assertEqual(list(problem.objects(Location)), [l1, l2, l3])
        self.assertEqual(problem.user_type("Robot"), Robot)
        self.assertEqual(len(list(problem.objects(Robot))), 1)
        self.assertEqual(list(problem.objects(Robot)), [r1])
        self.assertEqual(problem.user_type("Container"), Container)
        self.assertEqual(len(list(problem.objects(Container))), 1)
        self.assertEqual(list(problem.objects(Container)), [c1])
        self.assertEqual(len(problem.actions), 3)
        self.assertEqual(problem.action("move"), move)
        self.assertEqual(problem.action("load"), load)
        self.assertEqual(problem.action("unload"), unload)
        self.assertTrue(problem.initial_value(robot_at(r1, l1)) is not None)
        self.assertTrue(problem.initial_value(robot_at(r1, l2)) is not None)
        self.assertTrue(problem.initial_value(robot_at(r1, l3)) is not None)
        self.assertTrue(problem.initial_value(cargo_at(c1, l1)) is not None)
        self.assertTrue(problem.initial_value(cargo_at(c1, l2)) is not None)
        self.assertTrue(problem.initial_value(cargo_at(c1, l3)) is not None)
        self.assertTrue(problem.initial_value(cargo_mounted(c1, r1)) is not None)
        self.assertEqual(len(problem.goals), 2)

    def test_fluents_defaults(self):
        Location = UserType("Location")
        robot_at = Fluent("robot_at", BoolType(), position=Location)
        distance = Fluent(
            "distance", RealType(), location_1=Location, location_2=Location
        )

        N = 10
        locations = [Object(f"l{i}", Location) for i in range(N)]

        problem = Problem("robot")
        problem.add_fluent(robot_at, default_initial_value=False)
        problem.add_fluent(distance, default_initial_value=Fraction(-1))
        problem.add_objects(locations)
        problem.set_initial_value(robot_at(locations[0]), True)
        for i in range(N - 1):
            problem.set_initial_value(
                distance(locations[i], locations[i + 1]), Fraction(10)
            )

        self.assertEqual(problem.initial_value(robot_at(locations[0])), TRUE())
        for i in range(1, N):
            self.assertEqual(problem.initial_value(robot_at(locations[i])), FALSE())

        for i in range(N):
            for j in range(N):
                if j == i + 1:
                    self.assertEqual(
                        problem.initial_value(distance(locations[i], locations[j])),
                        Real(Fraction(10)),
                    )
                else:
                    self.assertEqual(
                        problem.initial_value(distance(locations[i], locations[j])),
                        Real(Fraction(-1)),
                    )

    def test_problem_defaults(self):
        Location = UserType("Location")
        robot_at = Fluent("robot_at", BoolType(), position=Location)
        distance = Fluent(
            "distance", IntType(), location_1=Location, location_2=Location
        )
        cost = Fluent("cost", IntType(), location_1=Location, location_2=Location)

        N = 10
        locations = [Object(f"l{i}", Location) for i in range(N)]

        problem = Problem("robot", initial_defaults={IntType(): 0})
        problem.add_fluent(robot_at, default_initial_value=False)
        problem.add_fluent(distance, default_initial_value=-1)
        problem.add_fluent(cost)
        problem.add_objects(locations)
        problem.set_initial_value(robot_at(locations[0]), True)
        for i in range(N - 1):
            problem.set_initial_value(distance(locations[i], locations[i + 1]), 10)
            problem.set_initial_value(cost(locations[i], locations[i + 1]), 100)

        self.assertEqual(problem.initial_value(robot_at(locations[0])), TRUE())
        for i in range(1, N):
            self.assertEqual(problem.initial_value(robot_at(locations[i])), FALSE())

        for i in range(N):
            for j in range(N):
                if j == i + 1:
                    self.assertEqual(
                        problem.initial_value(distance(locations[i], locations[j])),
                        Int(10),
                    )
                    self.assertEqual(
                        problem.initial_value(cost(locations[i], locations[j])),
                        Int(100),
                    )
                else:
                    self.assertEqual(
                        problem.initial_value(distance(locations[i], locations[j])),
                        Int(-1),
                    )
                    self.assertEqual(
                        problem.initial_value(cost(locations[i], locations[j])), Int(0)
                    )

    def test_htn_problem_creation(self):
        problems = examples.hierarchical.get_example_problems()
        problem = problems["htn-go"].problem
        self.assertTrue(isinstance(problem, up.model.htn.HierarchicalProblem))
        self.assertTrue(problem.kind.has_hierarchical())
        self.assertEqual(2, len(problem.fluents))
        self.assertEqual(1, len(problem.actions))
        self.assertEqual(["go"], [task.name for task in problem.tasks])
        self.assertEqual(
            ["go-noop", "go-recursive"], [method.name for method in problem.methods]
        )

        go_direct = problem.method("go-noop")
        self.assertEqual(0, len(go_direct.subtasks))
        self.assertEqual(1, len(go_direct.preconditions))
        self.assertEqual(0, len(go_direct.constraints))

        go_indirect = problem.method("go-recursive")
        self.assertEqual(2, len(go_indirect.subtasks))
        self.assertEqual(2, len(go_indirect.preconditions))
        self.assertEqual(1, len(go_indirect.constraints))

        self.assertEqual(2, len(problem.task_network.subtasks))


if __name__ == "__main__":
    main()



# Copyright 2022 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from dataclasses import dataclass
from typing import IO


@dataclass
class Credits:
    name: str
    author: str
    contact: str
    website: str
    license: str
    short_description: str
    long_description: str

    def write_credits(self, stream: IO[str], full_credits: bool = False):
        stream.write(f"  * Engine name: {self.name}\n  * Developers:  {self.author}\n")
        if not full_credits:
            stream.write("  * Description: ")
            stream.write(self.short_description.replace("\n", "\n  *              "))
            stream.write("\n")
        else:
            stream.write(f"  * Contacts:    {self.contact}\n")
            stream.write(f"  * Website:     {self.website}\n")
            stream.write(f"  * License:     {self.license}\n  * Description: ")
            stream.write(self.long_description.replace("\n", "\n  *              "))
            stream.write("\n")



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unified_planning
from unified_planning.shortcuts import *
from unified_planning.test import TestCase, main
from unified_planning.test.examples import get_example_problems
from unified_planning.engines import SequentialPlanValidator, ValidationResultStatus
from unified_planning.environment import get_env


class TestProblem(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()

    def test_all(self):
        pv = SequentialPlanValidator(env=get_env())
        for p in self.problems.values():
            if not pv.supports(p.problem.kind):
                continue
            problem, plan = p.problem, p.plan
            if SequentialPlanValidator.supports(problem.kind):
                validation_result = pv.validate(problem, plan)
                self.assertEqual(validation_result.status, ValidationResultStatus.VALID)

    def test_all_from_factory(self):
        with PlanValidator(name="sequential_plan_validator") as pv:
            self.assertEqual(pv.name, "sequential_plan_validator")
            for p in self.problems.values():
                if not pv.supports(p.problem.kind):
                    continue
                problem, plan = p.problem, p.plan
                validation_result = pv.validate(problem, plan)
                self.assertEqual(validation_result.status, ValidationResultStatus.VALID)

    def test_all_from_factory_with_problem_kind(self):
        for p in self.problems.values():
            problem, plan = p.problem, p.plan
            pk = problem.kind
            if SequentialPlanValidator.supports(pk):
                env = unified_planning.environment.Environment()
                env.factory.preference_list = [
                    e for e in env.factory.preference_list if e != "tamer"
                ]
                with env.factory.PlanValidator(
                    problem_kind=pk, plan_kind=plan.kind
                ) as pv:
                    self.assertEqual(pv.name, "sequential_plan_validator")
                    validation_result = pv.validate(problem, plan)
                    self.assertEqual(
                        validation_result.status, ValidationResultStatus.VALID
                    )



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
import unified_planning as up
from unified_planning.shortcuts import *
from unified_planning.engines.results import POSITIVE_OUTCOMES
from unified_planning.test import TestCase, main, skipIfEngineNotAvailable
from unified_planning.environment import get_env
from unified_planning.test.examples import get_example_problems


class TestPyperplan(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.env = get_env()
        self.problems = get_example_problems()

    @skipIfEngineNotAvailable("pyperplan")
    def test_pyperplan(self):
        problem, plan = self.problems["robot_no_negative_preconditions"]
        with OneshotPlanner(name="pyperplan") as planner:
            self.assertNotEqual(planner, None)
            final_report = planner.solve(problem)
            self.assertIn(final_report.status, POSITIVE_OUTCOMES)
            self.assertEqual(str(plan), str(final_report.plan))

    @skipIfEngineNotAvailable("pyperplan")
    def test_basic_without_negative_preconditions(self):
        problem, plan = self.problems["basic_without_negative_preconditions"]
        with OneshotPlanner(name="pyperplan") as planner:
            self.assertNotEqual(planner, None)
            final_report = planner.solve(problem)
            self.assertIn(final_report.status, POSITIVE_OUTCOMES)
            self.assertEqual(str(plan), str(final_report.plan))

    @skipIfEngineNotAvailable("pyperplan")
    def test_basic_nested_conjunctions(self):
        problem, plan = self.problems["basic_nested_conjunctions"]
        with OneshotPlanner(name="pyperplan") as planner:
            self.assertNotEqual(planner, None)
            final_report = planner.solve(problem)
            self.assertIn(final_report.status, POSITIVE_OUTCOMES)
            self.assertEqual(str(plan), str(final_report.plan))

    @skipIfEngineNotAvailable("pyperplan")
    def test_hierarchical_blocks_world(self):
        problem, plan = self.problems["hierarchical_blocks_world"]
        with OneshotPlanner(name="pyperplan") as planner:
            self.assertNotEqual(planner, None)
            final_report = planner.solve(problem)
            self.assertIn(final_report.status, POSITIVE_OUTCOMES)
            self.assertEqual(str(plan), str(final_report.plan))

    @skipIfEngineNotAvailable("pyperplan")
    def test_hierarchical_blocks_world_object_as_root(self):
        problem, plan = self.problems["hierarchical_blocks_world_object_as_root"]
        with OneshotPlanner(name="pyperplan") as planner:
            self.assertNotEqual(planner, None)
            final_report = planner.solve(problem)
            self.assertIn(final_report.status, POSITIVE_OUTCOMES)
            self.assertEqual(str(plan), str(final_report.plan))

    @skipIfEngineNotAvailable("pyperplan")
    def test_hierarchical_blocks_world_with_object(self):
        problem, plan = self.problems["hierarchical_blocks_world_with_object"]
        with OneshotPlanner(name="pyperplan") as planner:
            self.assertNotEqual(planner, None)
            final_report = planner.solve(problem)
            self.assertIn(final_report.status, POSITIVE_OUTCOMES)
            self.assertEqual(str(plan), str(final_report.plan))

    @skipIfEngineNotAvailable("pyperplan")
    def test_hierarchical_blocks_world_with_object_with_timeout(self):
        problem, plan = self.problems["hierarchical_blocks_world_with_object"]
        with OneshotPlanner(name="pyperplan") as planner:
            self.assertNotEqual(planner, None)
            with warnings.catch_warnings(record=True) as w:
                final_report = planner.solve(problem, timeout=0.001)
                self.assertIn(final_report.status, POSITIVE_OUTCOMES)
                self.assertEqual(str(plan), str(final_report.plan))
                self.assertEqual(len(w), 1)
                self.assertEqual(
                    "Pyperplan does not support timeout.", str(w[-1].message)
                )



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module defines the Fluent class.
A Fluent has a name, a type and a signature
that defines the types of its parameters.
"""

import unified_planning as up
from unified_planning.environment import get_env, Environment
from typing import List, OrderedDict, Optional, Union


class Fluent:
    """Represents a fluent."""

    def __init__(
        self,
        name: str,
        typename: "up.model.types.Type" = None,
        _signature: Optional[
            Union[
                OrderedDict[str, "up.model.types.Type"],
                List["up.model.parameter.Parameter"],
            ]
        ] = None,
        env: Environment = None,
        **kwargs: "up.model.types.Type",
    ):
        self._env = get_env(env)
        self._name = name
        if typename is None:
            self._typename = self._env.type_manager.BoolType()
        else:
            assert self._env.type_manager.has_type(
                typename
            ), "type of parameter does not belong to the same environment of the fluent"
            self._typename = typename
        self._signature: List["up.model.parameter.Parameter"] = []
        if _signature is not None:
            assert len(kwargs) == 0
            if isinstance(_signature, OrderedDict):
                for param_name, param_type in _signature.items():
                    self._signature.append(
                        up.model.parameter.Parameter(param_name, param_type, self._env)
                    )
            elif isinstance(_signature, List):
                assert all(
                    p.environment == self._env for p in _signature
                ), "one of the parameters does not belong to the same environment of the fluent"
                self._signature = _signature[:]
            else:
                raise NotImplementedError
        else:
            for param_name, param_type in kwargs.items():
                self._signature.append(
                    up.model.parameter.Parameter(param_name, param_type, self._env)
                )

    def __repr__(self) -> str:
        sign = ""
        if self.arity > 0:
            sign_items = [f"{p.name}={str(p.type)}" for p in self.signature]
            sign = f'[{", ".join(sign_items)}]'
        return f"{str(self.type)} {self.name}{sign}"

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, Fluent):
            return (
                self._name == oth._name
                and self._typename == oth._typename
                and self._signature == oth._signature
                and self._env == oth._env
            )
        else:
            return False

    def __hash__(self) -> int:
        res = hash(self._typename)
        for p in self._signature:
            res += hash(p)
        return res ^ hash(self._name)

    @property
    def name(self) -> str:
        """Returns the fluent name."""
        return self._name

    @property
    def type(self) -> "up.model.types.Type":
        """Returns the fluent type."""
        return self._typename

    @property
    def signature(self) -> List["up.model.parameter.Parameter"]:
        """Returns the fluent signature.
        The signature is the List of Parameters.
        """
        return self._signature

    @property
    def arity(self) -> int:
        """Returns the fluent arity.

        IMPORTANT NOTE: this property does some computation, so it should be called as
        seldom as possible."""
        return len(self._signature)

    @property
    def environment(self) -> "Environment":
        """Returns the Fluent environment."""
        return self._env

    def __call__(
        self, *args: "up.model.expression.Expression"
    ) -> "up.model.fnode.FNode":
        """Returns a fluent expression with the given parameters."""
        return self._env.expression_manager.FluentExp(self, args)

    #
    # Infix operators
    #

    def __add__(self, right):
        return self._env.expression_manager.Plus(self, right)

    def __radd__(self, left):
        return self._env.expression_manager.Plus(left, self)

    def __sub__(self, right):
        return self._env.expression_manager.Minus(self, right)

    def __rsub__(self, left):
        return self._env.expression_manager.Minus(left, self)

    def __mul__(self, right):
        return self._env.expression_manager.Times(self, right)

    def __rmul__(self, left):
        return self._env.expression_manager.Times(left, self)

    def __truediv__(self, right):
        return self._env.expression_manager.Div(self, right)

    def __rtruediv__(self, left):
        return self._env.expression_manager.Div(left, self)

    def __floordiv__(self, right):
        return self._env.expression_manager.Div(self, right)

    def __rfloordiv__(self, left):
        return self._env.expression_manager.Div(left, self)

    def __gt__(self, right):
        return self._env.expression_manager.GT(self, right)

    def __ge__(self, right):
        return self._env.expression_manager.GE(self, right)

    def __lt__(self, right):
        return self._env.expression_manager.LT(self, right)

    def __le__(self, right):
        return self._env.expression_manager.LE(self, right)

    def __pos__(self):
        return self._env.expression_manager.Plus(0, self)

    def __neg__(self):
        return self._env.expression_manager.Minus(0, self)

    def Equals(self, right):
        return self._env.expression_manager.Equals(self, right)

    def And(self, *other):
        return self._env.expression_manager.And(self, *other)

    def __and__(self, *other):
        return self._env.expression_manager.And(self, *other)

    def __rand__(self, *other):
        return self._env.expression_manager.And(*other, self)

    def Or(self, *other):
        return self._env.expression_manager.Or(self, *other)

    def __or__(self, *other):
        return self._env.expression_manager.Or(self, *other)

    def __ror__(self, *other):
        return self._env.expression_manager.Or(*other, self)

    def Not(self):
        return self._env.expression_manager.Not(self)

    def __invert__(self):
        return self._env.expression_manager.Not(self)

    def Xor(self, *other):
        em = self._env.expression_manager
        return em.And(em.Or(self, *other), em.Not(em.And(self, *other)))

    def __xor__(self, *other):
        em = self._env.expression_manager
        return em.And(em.Or(self, *other), em.Not(em.And(self, *other)))

    def __rxor__(self, other):
        em = self._env.expression_manager
        return em.And(em.Or(*other, self), em.Not(em.And(*other, self)))

    def Implies(self, right):
        return self._env.expression_manager.Implies(self, right)

    def Iff(self, right):
        return self._env.expression_manager.Iff(self, right)
